# --------------------------------------------------------------------

class ISType(object):
    """The index set types."""
    GENERAL = S_(ISGENERAL)
    BLOCK   = S_(ISBLOCK)
    STRIDE  = S_(ISSTRIDE)

# --------------------------------------------------------------------


cdef class IS(Object):
    """A collection of indices.

    IS objects are used to index into vectors and matrices and to set up vector
    scatters.

    See Also
    --------
    petsc.IS

    """

    Type = ISType

    #

    def __cinit__(self):
        self.obj = <PetscObject*> &self.iset
        self.iset = NULL

    # buffer interface (PEP 3118)

    def __getbuffer__(self, Py_buffer *view, int flags):
        cdef _IS_buffer buf = _IS_buffer(self)
        buf.acquirebuffer(view, flags)

    def __releasebuffer__(self, Py_buffer *view):
        cdef _IS_buffer buf = <_IS_buffer>(view.obj)
        buf.releasebuffer(view)
        <void>self # unused

    # 'with' statement (PEP 343)

    def __enter__(self):
        cdef _IS_buffer buf = _IS_buffer(self)
        self.set_attr('__buffer__', buf)
        return buf.enter()

    def __exit__(self, *exc):
        cdef _IS_buffer buf = self.get_attr('__buffer__')
        self.set_attr('__buffer__', None)
        return buf.exit()
    #

    def view(self, Viewer viewer=None) -> None:
        """Display the index set.

        Collective.

        Parameters
        ----------
        viewer
            Viewer used to display the IS.

        See Also
        --------
        petsc.ISView

        """
        cdef PetscViewer cviewer = NULL
        if viewer is not None: cviewer = viewer.vwr
        CHKERR(ISView(self.iset, cviewer))

    def destroy(self) -> Self:
        """Destroy the index set.

        Collective.

        See Also
        --------
        petsc.ISDestroy

        """
        CHKERR(ISDestroy(&self.iset))
        return self

    def create(self, comm: Comm | None = None) -> Self:
        """Create an IS.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.ISCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscIS newiset = NULL
        CHKERR(ISCreate(ccomm, &newiset))
        CHKERR(PetscCLEAR(self.obj)); self.iset = newiset
        return self

    def setType(self, is_type: IS.Type | str) -> None:
        """Set the type of the index set.

        Collective.

        Parameters
        ----------
        is_type
            The index set type.

        See Also
        --------
        petsc.ISSetType

        """
        cdef PetscISType cval = NULL
        is_type = str2bytes(is_type, &cval)
        CHKERR(ISSetType(self.iset, cval))

    def getType(self) -> str:
        """Return the index set type associated with the IS.

        Not collective.

        See Also
        --------
        petsc.ISGetType

        """
        cdef PetscISType cval = NULL
        CHKERR(ISGetType(self.iset, &cval))
        return bytes2str(cval)

    def createGeneral(
        self,
        indices: Sequence[int],
        comm: Comm | None = None) -> Self:
        """Create an IS with indices.

        Collective.

        Parameters
        ----------
        indices
            Integer array.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.ISCreateGeneral

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscInt nidx = 0, *idx = NULL
        cdef PetscCopyMode cm = PETSC_COPY_VALUES
        cdef PetscIS newiset = NULL
        indices = iarray_i(indices, &nidx, &idx)
        CHKERR(ISCreateGeneral(ccomm, nidx, idx, cm, &newiset))
        CHKERR(PetscCLEAR(self.obj)); self.iset = newiset
        return self

    def createBlock(
        self,
        bsize: int,
        indices: Sequence[int],
        comm: Comm | None = None) -> Self:
        """Create a blocked index set.

        Collective.

        Parameters
        ----------
        bsize
            Block size.
        indices
            Integer array of indices.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.ISCreateBlock

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscInt bs = asInt(bsize)
        cdef PetscInt nidx = 0, *idx = NULL
        cdef PetscCopyMode cm = PETSC_COPY_VALUES
        cdef PetscIS newiset = NULL
        indices = iarray_i(indices, &nidx, &idx)
        CHKERR(ISCreateBlock(ccomm, bs, nidx, idx, cm, &newiset))
        CHKERR(PetscCLEAR(self.obj)); self.iset = newiset
        return self

    def createStride(
        self,
        size: int,
        first: int = 0,
        step: int = 1,
        comm: Comm | None = None) -> Self:
        """Create an index set consisting of evenly spaced values.

        Collective.

        Parameters
        ----------
        size
            The length of the locally owned portion of the index set.
        first
            The first element of the index set.
        step
            The difference between adjacent indices.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.ISCreateStride

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscInt csize  = asInt(size)
        cdef PetscInt cfirst = asInt(first)
        cdef PetscInt cstep  = asInt(step)
        cdef PetscIS newiset = NULL
        CHKERR(ISCreateStride(ccomm, csize, cfirst, cstep, &newiset))
        CHKERR(PetscCLEAR(self.obj)); self.iset = newiset
        return self

    def duplicate(self) -> IS:
        """Create a copy of the index set.

        Collective.

        See Also
        --------
        IS.copy, petsc.ISDuplicate

        """
        cdef IS iset = type(self)()
        CHKERR(ISDuplicate(self.iset, &iset.iset))
        return iset

    def copy(self, IS result=None) -> IS:
        """Copy the contents of the index set into another.

        Collective.

        Parameters
        ----------
        result
            The target index set. If `None` then `IS.duplicate` is called first.

        Returns
        -------
        IS
            The copied index set. If ``result`` is not `None` then this is
            returned here.

        See Also
        --------
        IS.duplicate, petsc.ISCopy

        """
        if result is None:
            result = type(self)()
        if result.iset == NULL:
            CHKERR(ISDuplicate(self.iset, &result.iset))
        CHKERR(ISCopy(self.iset, result.iset))
        return result

    def load(self, Viewer viewer) -> Self:
        """Load a stored index set.

        Collective.

        Parameters
        ----------
        viewer
            Binary file viewer, either `Viewer.Type.BINARY` or `Viewer.Type.HDF5`.

        See Also
        --------
        petsc.ISLoad

        """
        cdef MPI_Comm comm = MPI_COMM_NULL
        cdef PetscObject obj = <PetscObject>(viewer.vwr)
        if self.iset == NULL:
            CHKERR(PetscObjectGetComm(obj, &comm))
            CHKERR(ISCreate(comm, &self.iset))
        CHKERR(ISLoad(self.iset, viewer.vwr))
        return self

    def allGather(self) -> IS:
        """Concatenate index sets stored across processors.

        Collective.

        The returned index set will be the same on every processor.

        See Also
        --------
        petsc.ISAllGather

        """
        cdef IS iset = IS()
        CHKERR(ISAllGather(self.iset, &iset.iset))
        return iset

    def toGeneral(self) -> Self:
        """Convert the index set type to `IS.Type.GENERAL`.

        Collective.

        See Also
        --------
        petsc.ISToGeneral, petsc.ISType

        """
        CHKERR(ISToGeneral(self.iset))
        return self

    def buildTwoSided(self, IS toindx=None) -> IS:
        """Create an index set describing a global mapping.

        Collective.

        This function generates an index set that contains new numbers from
        remote or local on the index set.

        Parameters
        ----------
        toindx
            Index set describing which indices to send, default is to send
            natural numbering.

        Returns
        -------
        IS
            New index set containing the new numbers from remote or local.

        See Also
        --------
        petsc.ISBuildTwoSided

        """
        cdef PetscIS ctoindx = NULL
        if toindx is not None: ctoindx = toindx.iset
        cdef IS result = IS()
        CHKERR(ISBuildTwoSided(self.iset, ctoindx, &result.iset))
        return result

    def invertPermutation(self, nlocal: int | None = None) -> IS:
        """Invert the index set.

        Collective.

        For this to be correct the index set must be a permutation.

        Parameters
        ----------
        nlocal
            The number of indices on this processor in the resulting index set,
            defaults to ``PETSC_DECIDE``.

        See Also
        --------
        petsc.ISInvertPermutation

        """
        cdef PetscInt cnlocal = PETSC_DECIDE
        if nlocal is not None: cnlocal = asInt(nlocal)
        cdef IS iset = IS()
        CHKERR(ISInvertPermutation(self.iset, cnlocal, &iset.iset))
        return iset

    def partitioningToNumbering(self) -> IS:
        """Return the new global numbering after a partitioning.

        Collective.

        Assuming that the index set represents a partitioning, generate another
        index set with the new global node number in the new ordering.

        See Also
        --------
        petsc.ISPartitioningToNumbering

        """
        cdef IS iset = IS()
        CHKERR(ISPartitioningToNumbering(self.iset, &iset.iset))
        return iset

    def partitioningCount(self, npart: int | None = None) -> ArrayInt:
        """Return the number of elements per process after a partitioning.

        Collective.

        Assuming that the index set represents a partitioning, determine the
        number of elements on each (partition) rank.

        Parameters
        ----------
        npart
            The number of partitions,
            defaults to the size of the communicator.

        See Also
        --------
        petsc.ISPartitioningCount

        """
        cdef PetscInt size
        if npart is None: size = getCommSize(self.iset)
        else: size = asInt(npart)
        cdef PetscInt *counts = NULL
        CHKERR(PetscMalloc(<size_t>size*sizeof(PetscInt), &counts))
        CHKERR(ISPartitioningCount(self.iset, size, counts))
        cdef object ocounts = None
        try:
            ocounts = array_i(size, counts)
        finally:
            CHKERR(PetscFree(counts))
        return ocounts

    def getSize(self) -> int:
        """Return the global length of an index set.

        Not collective.

        See Also
        --------
        petsc.ISGetSize

        """
        cdef PetscInt N = 0
        CHKERR(ISGetSize(self.iset, &N))
        return toInt(N)

    def getLocalSize(self) -> int:
        """Return the process-local length of the index set.

        Not collective.

        See Also
        --------
        petsc.ISGetLocalSize

        """
        cdef PetscInt n = 0
        CHKERR(ISGetLocalSize(self.iset, &n))
        return toInt(n)

    def getSizes(self) -> tuple[int, int]:
        """Return the local and global sizes of the index set.

        Not collective.

        Returns
        -------
        local_size : int
            The local size.
        global_size : int
            The global size.

        See Also
        --------
        IS.getLocalSize, IS.getSize

        """
        cdef PetscInt n = 0, N = 0
        CHKERR(ISGetLocalSize(self.iset, &n))
        CHKERR(ISGetSize(self.iset, &N))
        return (toInt(n), toInt(N))

    def getBlockSize(self) -> int:
        """Return the number of elements in a block.

        Not collective.

        See Also
        --------
        petsc.ISGetBlockSize

        """
        cdef PetscInt bs = 1
        CHKERR(ISGetBlockSize(self.iset, &bs))
        return toInt(bs)

    def setBlockSize(self, bs: int) -> None:
        """Set the block size of the index set.

        Logically collective.

        Parameters
        ----------
        bs
            Block size.

        See Also
        --------
        petsc.ISSetBlockSize

        """
        cdef PetscInt cbs = asInt(bs)
        CHKERR(ISSetBlockSize(self.iset, cbs))

    def sort(self) -> Self:
        """Sort the indices of an index set.

        Collective.

        See Also
        --------
        petsc.ISSort

        """
        CHKERR(ISSort(self.iset))
        return self

    def isSorted(self) -> bool:
        """Return whether the indices have been sorted.

        Collective.

        See Also
        --------
        petsc.ISSorted

        """
        cdef PetscBool flag = PETSC_FALSE
        CHKERR(ISSorted(self.iset, &flag))
        return toBool(flag)

    def setPermutation(self) -> Self:
        """Mark the index set as being a permutation.

        Logically collective.

        See Also
        --------
        petsc.ISSetPermutation

        """
        CHKERR(ISSetPermutation(self.iset))
        return self

    def isPermutation(self) -> bool:
        """Return whether an index set has been declared to be a permutation.

        Logically collective.

        See Also
        --------
        petsc.ISPermutation

        """
        cdef PetscBool flag = PETSC_FALSE
        CHKERR(ISPermutation(self.iset, &flag))
        return toBool(flag)

    def setIdentity(self) -> Self:
        """Mark the index set as being an identity.

        Logically collective.

        See Also
        --------
        petsc.ISSetIdentity

        """
        CHKERR(ISSetIdentity(self.iset))
        return self

    def isIdentity(self) -> bool:
        """Return whether the index set has been declared as an identity.

        Collective.

        See Also
        --------
        petsc.ISIdentity

        """
        cdef PetscBool flag = PETSC_FALSE
        CHKERR(ISIdentity(self.iset, &flag))
        return toBool(flag)

    def equal(self, IS iset) -> bool:
        """Return whether the index sets have the same set of indices or not.

        Collective.

        Parameters
        ----------
        iset
            The index set to compare indices with.

        See Also
        --------
        petsc.ISEqual

        """
        cdef PetscBool flag = PETSC_FALSE
        CHKERR(ISEqual(self.iset, iset.iset, &flag))
        return toBool(flag)

    def sum(self, IS iset) -> IS:
        """Return the union of two (sorted) index sets.

        Collective.

        Parameters
        ----------
        iset
            The index set to compute the union with.

        See Also
        --------
        petsc.ISSum

        """
        cdef IS out = IS()
        CHKERR(ISSum(self.iset, iset.iset, &out.iset))
        return out

    def expand(self, IS iset) -> IS:
        """Return the union of two (possibly unsorted) index sets.

        Collective.

        To compute the union, `expand` concatenates the two index sets
        and removes any duplicates.

        Parameters
        ----------
        iset
            Index set to compute the union with.

        Returns
        -------
        IS
            The new, combined, index set.

        See Also
        --------
        petsc.ISExpand

        """
        cdef IS out = IS()
        CHKERR(ISExpand(self.iset, iset.iset, &out.iset))
        return out

    def union(self, IS iset) -> IS:
        """Return the union of two (possibly unsorted) index sets.

        Collective.

        This function will call either `petsc.ISSum` or `petsc.ISExpand` depending
        on whether or not the input sets are already sorted.

        Sequential only (as `petsc.ISSum` is sequential only).

        Parameters
        ----------
        iset
            Index set to compute the union with.

        Returns
        -------
        IS
            The new, combined, index set.

        See Also
        --------
        IS.expand, IS.sum

        """
        cdef PetscBool flag1=PETSC_FALSE, flag2=PETSC_FALSE
        CHKERR(ISSorted(self.iset, &flag1))
        CHKERR(ISSorted(iset.iset, &flag2))
        cdef IS out = IS()
        if flag1==PETSC_TRUE and flag2==PETSC_TRUE:
            CHKERR(ISSum(self.iset, iset.iset, &out.iset))
        else:
            CHKERR(ISExpand(self.iset, iset.iset, &out.iset))
        return out

    def difference(self, IS iset: IS) -> IS:
        """Return the difference between two index sets.

        Collective.

        Parameters
        ----------
        iset
            Index set to compute the difference with.

        Returns
        -------
        IS
            Index set representing the difference between ``self`` and ``iset``.

        See Also
        --------
        petsc.ISDifference

        """
        cdef IS out = IS()
        CHKERR(ISDifference(self.iset, iset.iset, &out.iset))
        return out

    def complement(self, nmin: int, nmax: int) -> IS:
        """Create a complement index set.

        Collective.

        The complement set of indices is all indices that are not
        in the provided set (and within the provided bounds).

        Parameters
        ----------
        nmin
            Minimum index that can be found in the local part of the complement
            index set.
        nmax
            One greater than the maximum index that can be found in the local
            part of the complement index set.

        Notes
        -----
        For a parallel index set, this will generate the local part of the
        complement on each process.

        To generate the entire complement (on each process) of a parallel
        index set, first call `IS.allGather` and then call this method.

        See Also
        --------
        IS.allGather, petsc.ISComplement

        """
        cdef PetscInt cnmin = asInt(nmin)
        cdef PetscInt cnmax = asInt(nmax)
        cdef IS out = IS()
        CHKERR(ISComplement(self.iset, cnmin, cnmax, &out.iset))
        return out

    def embed(self, IS iset, drop: bool) -> IS:
        """Embed ``self`` into ``iset``.

        Not collective.

        The embedding is performed by finding the locations in ``iset`` that
        have the same indices as ``self``.

        Parameters
        ----------
        iset
            The index set to embed into.
        drop
            Flag indicating whether to drop indices from ``self`` that are not
            in ``iset``.

        Returns
        -------
        IS
            The embedded index set.

        See Also
        --------
        petsc.ISEmbed

        """
        cdef PetscBool bval = drop
        cdef IS out = IS()
        CHKERR(ISEmbed(self.iset, iset.iset, bval, &out.iset))
        return out

    def renumber(self, IS mult=None) -> tuple[int, IS]:
        """Renumber the non-negative entries of an index set, starting from 0.

        Collective.

        Parameters
        ----------
        mult
            The multiplicity of each entry in ``self``, default implies a
            multiplicity of 1.

        Returns
        -------
        int
            One past the largest entry of the new index set.
        IS
            The renumbered index set.

        See Also
        --------
        petsc.ISRenumber

        """
        cdef PetscIS mlt = NULL
        if mult is not None: mlt = mult.iset
        cdef IS out = IS()
        cdef PetscInt n = 0
        CHKERR(ISRenumber(self.iset, mlt, &n, &out.iset))
        return (toInt(n), out)
    #

    def setIndices(self, indices: Sequence[int]) -> None:
        """Set the indices of an index set.

        Logically collective.

        The index set is assumed to be of type `IS.Type.GENERAL`.

        See Also
        --------
        petsc.ISGeneralSetIndices

        """
        cdef PetscInt nidx = 0, *idx = NULL
        cdef PetscCopyMode cm = PETSC_COPY_VALUES
        indices = iarray_i(indices, &nidx, &idx)
        CHKERR(ISGeneralSetIndices(self.iset, nidx, idx, cm))

    def getIndices(self) -> ArrayInt:
        """Return the indices of the index set.

        Not collective.

        See Also
        --------
        petsc.ISGetIndices

        """
        cdef PetscInt size = 0
        cdef const PetscInt *indices = NULL
        CHKERR(ISGetLocalSize(self.iset, &size))
        CHKERR(ISGetIndices(self.iset, &indices))
        cdef object oindices = None
        try:
            oindices = array_i(size, indices)
        finally:
            CHKERR(ISRestoreIndices(self.iset, &indices))
        return oindices

    def setBlockIndices(self, bsize: int, indices: Sequence[int]) -> None:
        """Set the indices for an index set with type `IS.Type.BLOCK`.

        Collective.

        Parameters
        ----------
        bsize
            Number of elements in each block.
        indices
            List of integers.

        See Also
        --------
        petsc.ISBlockSetIndices

        """
        cdef PetscInt bs = asInt(bsize)
        cdef PetscInt nidx = 0, *idx = NULL
        cdef PetscCopyMode cm = PETSC_COPY_VALUES
        indices = iarray_i(indices, &nidx, &idx)
        CHKERR(ISBlockSetIndices(self.iset, bs, nidx, idx, cm))

    def getBlockIndices(self) -> ArrayInt:
        """Return the indices of an index set with type `IS.Type.BLOCK`.

        Not collective.

        See Also
        --------
        petsc.ISBlockGetIndices

        """
        cdef PetscInt size = 0, bs = 1
        cdef const PetscInt *indices = NULL
        CHKERR(ISGetLocalSize(self.iset, &size))
        CHKERR(ISGetBlockSize(self.iset, &bs))
        CHKERR(ISBlockGetIndices(self.iset, &indices))
        cdef object oindices = None
        try:
            oindices = array_i(size//bs, indices)
        finally:
            CHKERR(ISBlockRestoreIndices(self.iset, &indices))
        return oindices

    def setStride(self, size: int, first: int = 0, step: int = 1) -> None:
        """Set the stride information for an index set with type `IS.Type.STRIDE`.

        Logically collective.

        Parameters
        ----------
        size
            Length of the locally owned portion of the index set.
        first
            First element of the index set.
        step
            Difference between adjacent indices.

        See Also
        --------
        petsc.ISStrideSetStride

        """
        cdef PetscInt csize = asInt(size)
        cdef PetscInt cfirst = asInt(first)
        cdef PetscInt cstep = asInt(step)
        CHKERR(ISStrideSetStride(self.iset, csize, cfirst, cstep))

    def getStride(self) -> tuple[int, int, int]:
        """Return size and stride information.

        Not collective.

        Returns
        -------
        size : int
            Length of the locally owned portion of the index set.
        first : int
            First element of the index set.
        step : int
            Difference between adjacent indices.

        See Also
        --------
        petsc.ISGetLocalSize, petsc.ISStrideGetInfo

        """
        cdef PetscInt size=0, first=0, step=0
        CHKERR(ISGetLocalSize(self.iset, &size))
        CHKERR(ISStrideGetInfo(self.iset, &first, &step))
        return (toInt(size), toInt(first), toInt(step))

    def getInfo(self) -> tuple[int, int]:
        """Return stride information for an index set with type `IS.Type.STRIDE`.

        Not collective.

        Returns
        -------
        first : int
            First element of the index set.
        step : int
            Difference between adjacent indices.

        See Also
        --------
        IS.getStride, petsc.ISStrideGetInfo

        """
        cdef PetscInt first = 0, step = 0
        CHKERR(ISStrideGetInfo(self.iset, &first, &step))
        return (toInt(first), toInt(step))

    #

    property permutation:
        """`True` if index set is a permutation, `False` otherwise.

        Logically collective.

        See Also
        --------
        IS.isPermutation

        """
        def __get__(self) -> bool:
            return self.isPermutation()

    property identity:
        """`True` if index set is an identity, `False` otherwise.

        Collective.

        See Also
        --------
        IS.isIdentity

        """
        def __get__(self) -> bool:
            return self.isIdentity()

    property sorted:
        """`True` if index set is sorted, `False` otherwise.

        Collective.

        See Also
        --------
        IS.isSorted

        """
        def __get__(self) -> bool:
            return self.isSorted()

    #

    property sizes:
        """The local and global sizes of the index set.

        Not collective.

        See Also
        --------
        IS.getSizes

        """
        def __get__(self) -> tuple[int, int]:
            return self.getSizes()

    property size:
        """The global size of the index set.

        Not collective.

        See Also
        --------
        IS.getSize

        """
        def __get__(self) -> int:
            return self.getSize()

    property local_size:
        """The local size of the index set.

        Not collective.

        See Also
        --------
        IS.getLocalSize

        """
        def __get__(self) -> int:
            return self.getLocalSize()

    property block_size:
        """The number of elements in a block.

        Not collective.

        See Also
        --------
        IS.getBlockSize

        """
        def __get__(self) -> int:
            return self.getBlockSize()

    property indices:
        """The indices of the index set.

        Not collective.

        See Also
        --------
        IS.getIndices

        """
        def __get__(self) -> ArrayInt:
            return self.getIndices()

    property array:
        """View of the index set as an array of integers.

        Not collective.

        """
        def __get__(self) -> ArrayInt:
            return asarray(self)

    # --- NumPy array interface (legacy) ---

    property __array_interface__:
        def __get__(self):
            cdef _IS_buffer buf = _IS_buffer(self)
            return buf.__array_interface__

# --------------------------------------------------------------------


class GLMapMode(object):
    """Enum describing mapping behavior when global indices are missing.

    MASK
        Give missing global indices a local index of -1.
    DROP
        Drop missing global indices.

    See Also
    --------
    petsc.ISGlobalToLocalMappingMode

    """
    MASK = PETSC_IS_GTOLM_MASK
    DROP = PETSC_IS_GTOLM_DROP


class LGMapType(object):
    """Local to global map types."""
    BASIC = S_(ISLOCALTOGLOBALMAPPINGBASIC)
    HASH  = S_(ISLOCALTOGLOBALMAPPINGHASH)


# --------------------------------------------------------------------

cdef class LGMap(Object):
    """Mapping from a local to a global ordering.

    See Also
    --------
    petsc.ISLocalToGlobalMapping

    """

    MapMode = GLMapMode

    Type = LGMapType
    #

    def __cinit__(self) -> None:
        self.obj = <PetscObject*> &self.lgm
        self.lgm = NULL

    def __call__(
        self,
        indices: Sequence[int],
        result: ArrayInt | None = None) -> None:
        """Convert a locally numbered list of integers to a global numbering.

        Not collective.

        Parameters
        ----------
        indices
            Input indices in local numbering.
        result
            Array to write the global numbering to. If `None` then a
            new array will be allocated.

        See Also
        --------
        IS.apply, petsc.ISLocalToGlobalMappingApply

        """
        self.apply(indices, result)

    #

    def setType(self, lgmap_type: LGMap.Type | str) -> None:
        """Set the type of the local-to-global map.

        Logically collective.

        Parameters
        ----------
        lgmap_type
            The type of the local-to-global mapping.

        Notes
        -----
        Use ``-islocaltoglobalmapping_type`` to set the type in the
        options database.

        See Also
        --------
        petsc_options, petsc.ISLocalToGlobalMappingSetType

        """
        cdef PetscISLocalToGlobalMappingType cval = NULL
        lgmap_type = str2bytes(lgmap_type, &cval)
        CHKERR(ISLocalToGlobalMappingSetType(self.lgm, cval))

    def setFromOptions(self) -> None:
        """Set mapping options from the options database.

        Not collective.

        See Also
        --------
        petsc_options, petsc.ISLocalToGlobalMappingSetFromOptions

        """
        CHKERR(ISLocalToGlobalMappingSetFromOptions(self.lgm))

    def view(self, Viewer viewer=None) -> None:
        """View the local-to-global mapping.

        Not collective.

        Parameters
        ----------
        viewer
            Viewer instance, defaults to an instance of `Viewer.Type.ASCII`.

        See Also
        --------
        load, petsc.ISLocalToGlobalMappingView

        """
        cdef PetscViewer cviewer = NULL
        if viewer is not None: cviewer = viewer.vwr
        CHKERR(ISLocalToGlobalMappingView(self.lgm, cviewer))

    def load(self, Viewer viewer) -> Self:
        """Load a local-to-global mapping.

        Collective.

        See Also
        --------
        view, petsc.ISLocalToGlobalMappingLoad

        """
        cdef MPI_Comm comm = MPI_COMM_NULL
        cdef PetscObject obj = <PetscObject>(viewer.vwr)
        if self.lgm == NULL:
            CHKERR(PetscObjectGetComm(obj, &comm))
            CHKERR(ISLocalToGlobalMappingCreate(comm, 1, 0, NULL, PETSC_USE_POINTER, &self.lgm))
        CHKERR(ISLocalToGlobalMappingLoad(self.lgm, viewer.vwr))
        return self

    def destroy(self) -> Self:
        """Destroy the local-to-global mapping.

        Not collective.

        See Also
        --------
        petsc.ISLocalToGlobalMappingDestroy

        """
        CHKERR(ISLocalToGlobalMappingDestroy(&self.lgm))
        return self

    def create(
        self,
        indices: Sequence[int],
        bsize: int | None = None,
        comm: Comm | None = None) -> Self:
        """Create a local-to-global mapping.

        Not collective.

        Parameters
        ----------
        indices
            Global index for each local element.
        bsize
            Block size, defaults to 1.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.ISLocalToGlobalMappingCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscInt bs = 1, nidx = 0, *idx = NULL
        cdef PetscCopyMode cm = PETSC_COPY_VALUES
        cdef PetscLGMap newlgm = NULL
        if bsize is not None: bs = asInt(bsize)
        if bs == PETSC_DECIDE: bs = 1
        indices = iarray_i(indices, &nidx, &idx)
        CHKERR(ISLocalToGlobalMappingCreate(
                ccomm, bs, nidx, idx, cm, &newlgm))
        CHKERR(PetscCLEAR(self.obj)); self.lgm = newlgm
        return self

    def createIS(self, IS iset) -> Self:
        """Create a local-to-global mapping from an index set.

        Not collective.

        Parameters
        ----------
        iset
            Index set containing the global numbers for each local number.

        See Also
        --------
        petsc.ISLocalToGlobalMappingCreateIS

        """
        cdef PetscLGMap newlgm = NULL
        CHKERR(ISLocalToGlobalMappingCreateIS(
            iset.iset, &newlgm))
        CHKERR(PetscCLEAR(self.obj)); self.lgm = newlgm
        return self

    def createSF(self, SF sf, start: int) -> Self:
        """Create a local-to-global mapping from a star forest.

        Collective.

        Parameters
        ----------
        sf
            Star forest mapping contiguous local indices to (rank, offset).
        start
            First global index on this process.

        See Also
        --------
        petsc.ISLocalToGlobalMappingCreateSF

        """
        cdef PetscLGMap newlgm = NULL
        cdef PetscInt cstart = asInt(start)
        CHKERR(ISLocalToGlobalMappingCreateSF(sf.sf, cstart, &newlgm))
        CHKERR(PetscCLEAR(self.obj)); self.lgm = newlgm
        return self

    def getSize(self) -> int:
        """Return the local size of the local-to-global mapping.

        Not collective.

        See Also
        --------
        petsc.ISLocalToGlobalMappingGetSize

        """
        cdef PetscInt n = 0
        CHKERR(ISLocalToGlobalMappingGetSize(self.lgm, &n))
        return toInt(n)

    def getBlockSize(self) -> int:
        """Return the block size of the local-to-global mapping.

        Not collective.

        See Also
        --------
        petsc.ISLocalToGlobalMappingGetBlockSize

        """
        cdef PetscInt bs = 1
        CHKERR(ISLocalToGlobalMappingGetBlockSize(self.lgm, &bs))
        return toInt(bs)

    def getIndices(self) -> ArrayInt:
        """Return the global indices for each local point in the mapping.

        Not collective.

        See Also
        --------
        petsc.ISLocalToGlobalMappingGetIndices

        """
        cdef PetscInt size = 0
        cdef const PetscInt *indices = NULL
        CHKERR(ISLocalToGlobalMappingGetSize(
                self.lgm, &size))
        CHKERR(ISLocalToGlobalMappingGetIndices(
                self.lgm, &indices))
        cdef object oindices = None
        try:
            oindices = array_i(size, indices)
        finally:
            CHKERR(ISLocalToGlobalMappingRestoreIndices(
                    self.lgm, &indices))
        return oindices

    def getBlockIndices(self) -> ArrayInt:
        """Return the global indices for each local block.

        Not collective.

        See Also
        --------
        petsc.ISLocalToGlobalMappingGetBlockIndices

        """
        cdef PetscInt size = 0, bs = 1
        cdef const PetscInt *indices = NULL
        CHKERR(ISLocalToGlobalMappingGetSize(
                self.lgm, &size))
        CHKERR(ISLocalToGlobalMappingGetBlockSize(
                self.lgm, &bs))
        CHKERR(ISLocalToGlobalMappingGetBlockIndices(
                self.lgm, &indices))
        cdef object oindices = None
        try:
            oindices = array_i(size//bs, indices)
        finally:
            CHKERR(ISLocalToGlobalMappingRestoreBlockIndices(
                    self.lgm, &indices))
        return oindices

    def getInfo(self) -> dict[int, ArrayInt]:
        """Determine the indices shared with neighboring processes.

        Collective.

        Returns
        -------
        dict
            Mapping from neighboring processor number to an array of shared
            indices (in local numbering).

        See Also
        --------
        petsc.ISLocalToGlobalMappingGetInfo

        """
        cdef PetscInt i, nproc = 0, *procs = NULL,
        cdef PetscInt *numprocs = NULL, **indices = NULL
        cdef dict neighs = {}
        CHKERR(ISLocalToGlobalMappingGetInfo(
                self.lgm, &nproc, &procs, &numprocs, &indices))
        try:
            for i from 0 <= i < nproc:
                neighs[toInt(procs[i])] = array_i(numprocs[i], indices[i])
        finally:
            ISLocalToGlobalMappingRestoreInfo(
                self.lgm, &nproc, &procs, &numprocs, &indices)
        return neighs

    def getBlockInfo(self) -> dict[int, ArrayInt]:
        """Determine the block indices shared with neighboring processes.

        Collective.

        Returns
        -------
        dict
            Mapping from neighboring processor number to an array of shared
            block indices (in local numbering).

        See Also
        --------
        petsc.ISLocalToGlobalMappingGetBlockInfo

        """
        cdef PetscInt i, nproc = 0, *procs = NULL,
        cdef PetscInt *numprocs = NULL, **indices = NULL
        cdef dict neighs = {}
        CHKERR(ISLocalToGlobalMappingGetBlockInfo(
                self.lgm, &nproc, &procs, &numprocs, &indices))
        try:
            for i from 0 <= i < nproc:
                neighs[toInt(procs[i])] = array_i(numprocs[i], indices[i])
        finally:
            ISLocalToGlobalMappingRestoreBlockInfo(
                self.lgm, &nproc, &procs, &numprocs, &indices)
        return neighs

    #

    def apply(
        self,
        indices: Sequence[int],
        result: ArrayInt | None = None) -> ArrayInt:
        """Convert a locally numbered list of integers to a global numbering.

        Not collective.

        Parameters
        ----------
        indices
            Input indices in local numbering.
        result
            Array to write the global numbering to. If `None` then a
            new array will be allocated.

        Returns
        -------
        ArrayInt
            Indices in global numbering. If ``result`` is not `None` then this is
            returned here.

        See Also
        --------
        LGMap.applyBlock, petsc.ISLocalToGlobalMappingApply

        """
        cdef PetscInt niidx = 0, *iidx = NULL
        cdef PetscInt noidx = 0, *oidx = NULL
        indices = iarray_i(indices, &niidx, &iidx)
        if result is None: result = empty_i(niidx)
        result  = oarray_i(result,  &noidx, &oidx)
        assert niidx == noidx, "incompatible array sizes"
        CHKERR(ISLocalToGlobalMappingApply(
            self.lgm, niidx, iidx, oidx))
        return result

    def applyBlock(
        self,
        indices: Sequence[int],
        result: ArrayInt | None = None) -> ArrayInt:
        """Convert a local block numbering to a global block numbering.

        Not collective.

        Parameters
        ----------
        indices
            Input block indices in local numbering.
        result
            Array to write the global numbering to. If `None` then a
            new array will be allocated.

        Returns
        -------
        ArrayInt
            Block indices in global numbering. If ``result`` is not `None`
            then this is returned here.

        See Also
        --------
        LGMap.apply, petsc.ISLocalToGlobalMappingApplyBlock

        """
        cdef PetscInt niidx = 0, *iidx = NULL
        cdef PetscInt noidx = 0, *oidx = NULL
        indices = iarray_i(indices, &niidx, &iidx)
        if result is None: result = empty_i(niidx)
        result  = oarray_i(result,  &noidx, &oidx)
        assert niidx == noidx, "incompatible array sizes"
        CHKERR(ISLocalToGlobalMappingApplyBlock(
            self.lgm, niidx, iidx, oidx))
        return result

    def applyIS(self, IS iset) -> IS:
        """Create an index set with global numbering from a local numbering.

        Collective.

        Parameters
        ----------
        iset
            Index set with local numbering.

        Returns
        -------
        IS
            Index set with global numbering.

        See Also
        --------
        petsc.ISLocalToGlobalMappingApplyIS

        """
        cdef IS result = IS()
        CHKERR(ISLocalToGlobalMappingApplyIS(
            self.lgm, iset.iset, &result.iset))
        return result

    def applyInverse(
        self,
        indices: Sequence[int],
        mode: GLMapMode | str | None = None) -> ArrayInt:
        """Compute local numbering from global numbering.

        Not collective.

        Parameters
        ----------
        indices
            Indices with a global numbering.
        mode
            Flag indicating what to do with indices that have no local value,
            defaults to ``"mask"``.

        Returns
        -------
        ArrayInt
            Indices with a local numbering.

        See Also
        --------
        petsc.ISGlobalToLocalMappingApply

        """
        cdef PetscGLMapMode cmode = PETSC_IS_GTOLM_MASK
        if mode is not None: cmode = mode
        cdef PetscInt n = 0, *idx = NULL
        indices = iarray_i(indices, &n, &idx)
        cdef PetscInt nout = n, *idxout = NULL
        if cmode != PETSC_IS_GTOLM_MASK:
            CHKERR(ISGlobalToLocalMappingApply(
                    self.lgm, cmode, n, idx, &nout, NULL))
        result = oarray_i(empty_i(nout), &nout, &idxout)
        CHKERR(ISGlobalToLocalMappingApply(
                self.lgm, cmode, n, idx, &nout, idxout))
        return result

    def applyBlockInverse(
        self,
        indices: Sequence[int],
        mode: GLMapMode | str | None = None) -> ArrayInt:
        """Compute blocked local numbering from blocked global numbering.

        Not collective.

        Parameters
        ----------
        indices
            Indices with a global block numbering.
        mode
            Flag indicating what to do with indices that have no local value,
            defaults to ``"mask"``.

        Returns
        -------
        ArrayInt
            Indices with a local block numbering.

        See Also
        --------
        petsc.ISGlobalToLocalMappingApplyBlock

        """
        cdef PetscGLMapMode cmode = PETSC_IS_GTOLM_MASK
        if mode is not None: cmode = mode
        cdef PetscInt n = 0, *idx = NULL
        indices = iarray_i(indices, &n, &idx)
        cdef PetscInt nout = n, *idxout = NULL
        if cmode != PETSC_IS_GTOLM_MASK:
            CHKERR(ISGlobalToLocalMappingApply(
                    self.lgm, cmode, n, idx, &nout, NULL))
        result = oarray_i(empty_i(nout), &nout, &idxout)
        CHKERR(ISGlobalToLocalMappingApplyBlock(
                self.lgm, cmode, n, idx, &nout, idxout))
        return result
    #

    property size:
        """The local size.

        Not collective.

        See Also
        --------
        LGMap.getSize

        """
        def __get__(self) -> int:
            return self.getSize()

    property block_size:
        """The block size.

        Not collective.

        See Also
        --------
        LGMap.getBlockSize

        """
        def __get__(self) -> int:
            return self.getBlockSize()

    property indices:
        """The global indices for each local point in the mapping.

        Not collective.

        See Also
        --------
        LGMap.getIndices, petsc.ISLocalToGlobalMappingGetIndices

        """
        def __get__(self) -> ArrayInt:
            return self.getIndices()

    property block_indices:
        """The global indices for each local block in the mapping.

        Not collective.

        See Also
        --------
        LGMap.getBlockIndices, petsc.ISLocalToGlobalMappingGetBlockIndices

        """
        def __get__(self) -> ArrayInt:
            return self.getBlockIndices()

    property info:
        """Mapping describing indices shared with neighboring processes.

        Collective.

        See Also
        --------
        LGMap.getInfo, petsc.ISLocalToGlobalMappingGetInfo

        """
        def __get__(self) -> dict[int, ArrayInt]:
            return self.getInfo()

    property block_info:
        """Mapping describing block indices shared with neighboring processes.

        Collective.

        See Also
        --------
        LGMap.getBlockInfo, petsc.ISLocalToGlobalMappingGetBlockInfo

        """
        def __get__(self) -> dict[int, ArrayInt]:
            return self.getBlockInfo()

# --------------------------------------------------------------------

del ISType
del GLMapMode
del LGMapType
# --------------------------------------------------------------------



# \begin{eqnarray}
#                  ys' = -2.0*\frac{-1.0+ys^2.0-\cos(t)}{2.0*ys}+0.05*\frac{-2.0+yf^2-\cos(5.0*t)}{2.0*yf}-\frac{\sin(t)}{2.0*ys}\\
#                  yf' = 0.05*\frac{-1.0+ys^2-\cos(t)}{2.0*ys}-\frac{-2.0+yf^2-\cos(5.0*t)}{2.0*yf}-5.0*\frac{\sin(5.0*t)}{2.0*yf}\\
# \end{eqnarray}

# This example demonstrates how to use ARKIMEX for solving a fast-slow system. The system is partitioned additively and component-wise at the same time.
# ys stands for the slow component and yf stands for the fast component. On the RHS for yf, only the term -\frac{-2.0+yf^2-\cos(5.0*t)}{2.0*yf} is treated implicitly while the rest is treated explicitly.

import sys
import petsc4py

petsc4py.init(sys.argv)

from petsc4py import PETSc
import numpy as np


class FSS:
    n = 2
    comm = PETSc.COMM_SELF

    def __init__(self):
        pass

    def initialCondition(self, u):
        u[0] = np.sqrt(2.0)
        u[1] = np.sqrt(3.0)
        u.assemble()

    def evalRHSFunction(self, ts, t, u, f):
        f[0] = (
            -2.0 * (-1.0 + u[0] * u[0] - np.cos(t)) / (2.0 * u[0])
            + 0.05 * (-2.0 + u[1] * u[1] - np.cos(5.0 * t)) / (2.0 * u[1])
            - np.sin(t) / (2.0 * u[0])
        )
        f[1] = (
            0.05 * (-1.0 + u[0] * u[0] - np.cos(t)) / (2.0 * u[0])
            - (-2.0 + u[1] * u[1] - np.cos(5.0 * t)) / (2.0 * u[1])
            - 5.0 * np.sin(5.0 * t) / (2.0 * u[1])
        )
        f.assemble()

    def evalRHSFunctionSlow(self, ts, t, u, f):
        f[0] = (
            -2.0 * (-1.0 + u[0] * u[0] - np.cos(t)) / (2.0 * u[0])
            + 0.05 * (-2.0 + u[1] * u[1] - np.cos(5.0 * t)) / (2.0 * u[1])
            - np.sin(t) / (2.0 * u[0])
        )
        f.assemble()

    def evalRHSFunctionFast(self, ts, t, u, f):
        f[0] = 0.05 * (-1.0 + u[0] * u[0] - np.cos(t)) / (2.0 * u[0]) - 5.0 * np.sin(
            5.0 * t
        ) / (2.0 * u[1])
        f.assemble()

    def evalIFunctionFast(self, ts, t, u, udot, f):
        f[0] = udot[0] + (-2.0 + u[1] * u[1] - np.cos(5.0 * t)) / (2.0 * u[1])
        f.assemble()

    def evalIJacobianFast(self, ts, t, u, udot, a, A, B):
        A[0, 0] = a + (2.0 + np.cos(5.0 * t)) / (2.0 * u[1] * u[1]) + 0.5
        A.assemble()
        if A != B:
            B.assemble()


OptDB = PETSc.Options()
explicitform_ = OptDB.getBool('explicitform', False)

ode = FSS()

Jim = PETSc.Mat().createDense([1, 1], comm=ode.comm)
Jim.setUp()

u = PETSc.Vec().createSeq(ode.n, comm=ode.comm)

ts = PETSc.TS().create(comm=ode.comm)
ts.setProblemType(ts.ProblemType.NONLINEAR)

if not explicitform_:
    iss = PETSc.IS().createGeneral([0], comm=ode.comm)
    isf = PETSc.IS().createGeneral([1], comm=ode.comm)
    ts.setType(ts.Type.ARKIMEX)
    ts.setARKIMEXFastSlowSplit(True)
    ts.setRHSSplitIS('slow', iss)
    ts.setRHSSplitIS('fast', isf)
    ts.setRHSSplitRHSFunction('slow', ode.evalRHSFunctionSlow, None)
    ts.setRHSSplitRHSFunction('fast', ode.evalRHSFunctionFast, None)
    ts.setRHSSplitIFunction('fast', ode.evalIFunctionFast, None)
    ts.setRHSSplitIJacobian('fast', ode.evalIJacobianFast, Jim, Jim)
else:
    f = u.duplicate()
    ts.setType(ts.Type.RK)
    ts.setRHSFunction(ode.evalRHSFunction, f)

ts.setTimeStep(0.01)
ts.setMaxTime(0.3)
ts.setExactFinalTime(PETSc.TS.ExactFinalTime.MATCHSTEP)

ts.setFromOptions()
ode.initialCondition(u)
ts.solve(u)
u.view()

del ode, Jim, u, ts



from petsc4py.PETSc import Object
from petsc4py.PETSc import Viewer


# A template class with the Python methods supported by PETSCVIEWERPYTHON


class PetscViewerPythonProtocol:
    def viewObject(self, viewer: Viewer, obj: Object) -> None:
        """View a generic object."""
        ...

    def setUp(self, viewer: Viewer) -> None:
        """Setup the viewer."""
        ...

    def setFromOptions(self, viewer: Viewer) -> None:
        """Process command line for customization."""
        ...

    def flush(self, viewer: Viewer) -> None:
        """Flush the viewer."""
        ...

    def view(self, viewer: Viewer, outviewer: Viewer) -> None:
        """View the viewer."""
        ...



cdef extern from * nogil:

    ctypedef const char* PetscRegressorType "PetscRegressorType"
    PetscRegressorType PETSCREGRESSORLINEAR

    PetscErrorCode PetscRegressorCreate(MPI_Comm, PetscRegressor*)
    PetscErrorCode PetscRegressorReset(PetscRegressor)
    PetscErrorCode PetscRegressorDestroy(PetscRegressor*)
    PetscErrorCode PetscRegressorSetType(PetscRegressor, PetscRegressorType)
    PetscErrorCode PetscRegressorGetType(PetscRegressor, PetscRegressorType*)
    PetscErrorCode PetscRegressorSetRegularizerWeight(PetscRegressor, PetscReal)
    PetscErrorCode PetscRegressorSetUp(PetscRegressor)
    PetscErrorCode PetscRegressorSetFromOptions(PetscRegressor)
    PetscErrorCode PetscRegressorView(PetscRegressor, PetscViewer)
    PetscErrorCode PetscRegressorFit(PetscRegressor, PetscMat, PetscVec)
    PetscErrorCode PetscRegressorPredict(PetscRegressor, PetscMat, PetscVec)
    PetscErrorCode PetscRegressorGetTao(PetscRegressor, PetscTAO*)

    PetscErrorCode PetscRegressorLinearSetFitIntercept(PetscRegressor, PetscBool)
    PetscErrorCode PetscRegressorLinearSetUseKSP(PetscRegressor, PetscBool)
    PetscErrorCode PetscRegressorLinearGetKSP(PetscRegressor, PetscKSP*)
    PetscErrorCode PetscRegressorLinearGetCoefficients(PetscRegressor, PetscVec*)
    PetscErrorCode PetscRegressorLinearGetIntercept(PetscRegressor, PetscScalar*)
    PetscErrorCode PetscRegressorLinearSetType(PetscRegressor, PetscRegressorLinearType)
    PetscErrorCode PetscRegressorLinearGetType(PetscRegressor, PetscRegressorLinearType*)

    ctypedef enum PetscRegressorLinearType:
        REGRESSOR_LINEAR_OLS
        REGRESSOR_LINEAR_LASSO
        REGRESSOR_LINEAR_RIDGE



# --------------------------------------------------------------------

cdef extern from * nogil:

    ctypedef const char* PetscVecType "VecType"
    PetscVecType VECSEQ
    PetscVecType VECMPI
    PetscVecType VECSTANDARD
    PetscVecType VECSHARED
    PetscVecType VECSEQVIENNACL
    PetscVecType VECMPIVIENNACL
    PetscVecType VECVIENNACL
    PetscVecType VECSEQCUDA
    PetscVecType VECMPICUDA
    PetscVecType VECCUDA
    PetscVecType VECSEQHIP
    PetscVecType VECMPIHIP
    PetscVecType VECHIP
    PetscVecType VECNEST
    PetscVecType VECSEQKOKKOS
    PetscVecType VECMPIKOKKOS
    PetscVecType VECKOKKOS

    ctypedef enum PetscVecOption "VecOption":
        VEC_IGNORE_OFF_PROC_ENTRIES
        VEC_IGNORE_NEGATIVE_INDICES

    PetscErrorCode VecView(PetscVec, PetscViewer)
    PetscErrorCode VecDestroy(PetscVec*)
    PetscErrorCode VecCreate(MPI_Comm, PetscVec*)

    PetscErrorCode VecSetOptionsPrefix(PetscVec, char[])
    PetscErrorCode VecAppendOptionsPrefix(PetscVec, char[])
    PetscErrorCode VecGetOptionsPrefix(PetscVec, char*[])
    PetscErrorCode VecSetFromOptions(PetscVec)
    PetscErrorCode VecSetUp(PetscVec)

    PetscErrorCode VecCreateSeq(MPI_Comm, PetscInt, PetscVec*)
    PetscErrorCode VecCreateSeqWithArray(MPI_Comm, PetscInt, PetscInt, PetscScalar[], PetscVec*)
    PetscErrorCode VecCreateSeqCUDAWithArrays(MPI_Comm, PetscInt, PetscInt, PetscScalar[], PetscScalar[], PetscVec*)
    PetscErrorCode VecCreateSeqHIPWithArrays(MPI_Comm, PetscInt, PetscInt, PetscScalar[], PetscScalar[], PetscVec*)
    PetscErrorCode VecCreateSeqViennaCLWithArrays(MPI_Comm, PetscInt, PetscInt, PetscScalar[], PetscScalar[], PetscVec*)
    PetscErrorCode VecCreateMPI(MPI_Comm, PetscInt, PetscInt, PetscVec*)
    PetscErrorCode VecCreateMPIWithArray(MPI_Comm, PetscInt, PetscInt, PetscInt, PetscScalar[], PetscVec*)
    PetscErrorCode VecCreateMPICUDAWithArrays(MPI_Comm, PetscInt, PetscInt, PetscInt, PetscScalar[], PetscScalar[], PetscVec*)
    PetscErrorCode VecCreateMPIHIPWithArrays(MPI_Comm, PetscInt, PetscInt, PetscInt, PetscScalar[], PetscScalar[], PetscVec*)
    PetscErrorCode VecCreateMPIViennaCLWithArrays(MPI_Comm, PetscInt, PetscInt, PetscInt, PetscScalar[], PetscScalar[], PetscVec*)
    PetscErrorCode VecCreateGhost(MPI_Comm, PetscInt, PetscInt, PetscInt, PetscInt[], PetscVec*)
    PetscErrorCode VecCreateGhostWithArray(MPI_Comm, PetscInt, PetscInt, PetscInt, PetscInt[], PetscScalar[], PetscVec*)
    PetscErrorCode VecCreateGhostBlock(MPI_Comm, PetscInt, PetscInt, PetscInt, PetscInt, PetscInt[], PetscVec*)
    PetscErrorCode VecCreateGhostBlockWithArray(MPI_Comm, PetscInt, PetscInt, PetscInt, PetscInt, PetscInt[], PetscScalar[], PetscVec*)
    PetscErrorCode VecCreateShared(MPI_Comm, PetscInt, PetscInt, PetscVec*)
    PetscErrorCode VecCreateNest(MPI_Comm, PetscInt, PetscIS[], PetscVec[], PetscVec*)
    PetscErrorCode VecGetType(PetscVec, PetscVecType*)
    PetscErrorCode VecSetType(PetscVec, PetscVecType)
    PetscErrorCode VecSetOption(PetscVec, PetscVecOption, PetscBool)
    PetscErrorCode VecSetSizes(PetscVec, PetscInt, PetscInt)
    PetscErrorCode VecGetSize(PetscVec, PetscInt*)
    PetscErrorCode VecGetLocalSize(PetscVec, PetscInt*)
    PetscErrorCode VecSetBlockSize(PetscVec, PetscInt)
    PetscErrorCode VecGetBlockSize(PetscVec, PetscInt*)
    PetscErrorCode VecGetOwnershipRange(PetscVec, PetscInt*, PetscInt*)
    PetscErrorCode VecGetOwnershipRanges(PetscVec, const PetscInt*[])

    PetscErrorCode VecCreateLocalVector(PetscVec, PetscVec*)
    PetscErrorCode VecGetLocalVector(PetscVec, PetscVec)
    PetscErrorCode VecRestoreLocalVector(PetscVec, PetscVec)
    PetscErrorCode VecGetLocalVectorRead(PetscVec, PetscVec)
    PetscErrorCode VecRestoreLocalVectorRead(PetscVec, PetscVec)

    PetscErrorCode VecGetArrayWrite(PetscVec, PetscScalar*[])
    PetscErrorCode VecRestoreArrayWrite(PetscVec, PetscScalar*[])
    PetscErrorCode VecGetArrayRead(PetscVec, const PetscScalar*[])
    PetscErrorCode VecRestoreArrayRead(PetscVec, const PetscScalar*[])
    PetscErrorCode VecGetArray(PetscVec, PetscScalar*[])
    PetscErrorCode VecRestoreArray(PetscVec, PetscScalar*[])
    PetscErrorCode VecPlaceArray(PetscVec, PetscScalar[])
    PetscErrorCode VecResetArray(PetscVec)
    PetscErrorCode VecGetArrayWriteAndMemType(PetscVec, PetscScalar*[], PetscMemType*)
    PetscErrorCode VecRestoreArrayWriteAndMemType(PetscVec, PetscScalar*[])
    PetscErrorCode VecGetArrayReadAndMemType(PetscVec, const PetscScalar*[], PetscMemType*)
    PetscErrorCode VecRestoreArrayReadAndMemType(PetscVec, const PetscScalar*[])
    PetscErrorCode VecGetArrayAndMemType(PetscVec, PetscScalar*[], PetscMemType*)
    PetscErrorCode VecRestoreArrayAndMemType(PetscVec, PetscScalar*[])

    PetscErrorCode VecEqual(PetscVec, PetscVec, PetscBool*)
    PetscErrorCode VecLoad(PetscVec, PetscViewer)

    PetscErrorCode VecDuplicate(PetscVec, PetscVec*)
    PetscErrorCode VecCopy(PetscVec, PetscVec)
    PetscErrorCode VecFilter(PetscVec, PetscReal)

    PetscErrorCode VecDuplicateVecs(PetscVec, PetscInt, PetscVec*[])
    PetscErrorCode VecDestroyVecs(PetscInt, PetscVec*[])

    PetscErrorCode VecGetValues(PetscVec, PetscInt, PetscInt[], PetscScalar[])

    PetscErrorCode VecSetValue(PetscVec, PetscInt, PetscScalar, PetscInsertMode)
    PetscErrorCode VecSetValues(PetscVec, PetscInt, const PetscInt[], const PetscScalar[], PetscInsertMode)
    PetscErrorCode VecSetValuesBlocked(PetscVec, PetscInt, const PetscInt[], const PetscScalar[], PetscInsertMode)

    PetscErrorCode VecSetLocalToGlobalMapping(PetscVec, PetscLGMap)
    PetscErrorCode VecGetLocalToGlobalMapping(PetscVec, PetscLGMap*)
    PetscErrorCode VecSetValueLocal(PetscVec, PetscInt, PetscScalar, PetscInsertMode)
    PetscErrorCode VecSetValuesLocal(PetscVec, PetscInt, const PetscInt[], const PetscScalar[], PetscInsertMode)
    PetscErrorCode VecSetValuesBlockedLocal(PetscVec, PetscInt, const PetscInt[], const PetscScalar[], PetscInsertMode)

    PetscErrorCode VecDot(PetscVec, PetscVec, PetscScalar*)
    PetscErrorCode VecDotBegin(PetscVec, PetscVec, PetscScalar*)
    PetscErrorCode VecDotEnd(PetscVec, PetscVec, PetscScalar*)
    PetscErrorCode VecTDot(PetscVec, PetscVec, PetscScalar*)
    PetscErrorCode VecTDotBegin(PetscVec, PetscVec, PetscScalar*)
    PetscErrorCode VecTDotEnd(PetscVec, PetscVec, PetscScalar*)
    PetscErrorCode VecMDot(PetscVec, PetscInt, PetscVec[], PetscScalar*)
    PetscErrorCode VecMDotBegin(PetscVec, PetscInt, PetscVec[], PetscScalar*)
    PetscErrorCode VecMDotEnd(PetscVec, PetscInt, PetscVec[], PetscScalar*)
    PetscErrorCode VecMTDot(PetscVec, PetscInt, PetscVec[], PetscScalar*)
    PetscErrorCode VecMTDotBegin(PetscVec, PetscInt, PetscVec[], PetscScalar*)
    PetscErrorCode VecMTDotEnd(PetscVec, PetscInt, PetscVec[], PetscScalar*)

    PetscErrorCode VecNorm(PetscVec, PetscNormType, PetscReal*)
    PetscErrorCode VecNormBegin(PetscVec, PetscNormType, PetscReal*)
    PetscErrorCode VecNormEnd(PetscVec, PetscNormType, PetscReal*)
    PetscErrorCode VecDotNorm2(PetscVec, PetscVec, PetscScalar*, PetscReal*)

    PetscErrorCode VecAssemblyBegin(PetscVec)
    PetscErrorCode VecAssemblyEnd(PetscVec)

    PetscErrorCode VecZeroEntries(PetscVec)
    PetscErrorCode VecConjugate(PetscVec)
    PetscErrorCode VecNormalize(PetscVec, PetscReal*)
    PetscErrorCode VecSum(PetscVec, PetscScalar*)
    PetscErrorCode VecMean(PetscVec, PetscScalar*)
    PetscErrorCode VecMax(PetscVec, PetscInt*, PetscReal*)
    PetscErrorCode VecMin(PetscVec, PetscInt*, PetscReal*)
    PetscErrorCode VecScale(PetscVec, PetscScalar)
    PetscErrorCode VecCopy(PetscVec, PetscVec)
    PetscErrorCode VecSetRandom(PetscVec, PetscRandom)
    PetscErrorCode VecSet(PetscVec, PetscScalar)
    PetscErrorCode VecSwap(PetscVec, PetscVec)
    PetscErrorCode VecAXPY(PetscVec, PetscScalar, PetscVec)
    PetscErrorCode VecAXPBY(PetscVec, PetscScalar, PetscScalar, PetscVec)
    PetscErrorCode VecAYPX(PetscVec, PetscScalar, PetscVec)
    PetscErrorCode VecWAXPY(PetscVec, PetscScalar, PetscVec, PetscVec)
    PetscErrorCode VecMAXPY(PetscVec, PetscInt, PetscScalar[], PetscVec[])
    PetscErrorCode VecPointwiseMax(PetscVec, PetscVec, PetscVec)
    PetscErrorCode VecPointwiseMaxAbs(PetscVec, PetscVec, PetscVec)
    PetscErrorCode VecPointwiseMin(PetscVec, PetscVec, PetscVec)
    PetscErrorCode VecPointwiseMult(PetscVec, PetscVec, PetscVec)
    PetscErrorCode VecPointwiseDivide(PetscVec, PetscVec, PetscVec)
    PetscErrorCode VecMaxPointwiseDivide(PetscVec, PetscVec, PetscReal*)
    PetscErrorCode VecShift(PetscVec, PetscScalar)
    PetscErrorCode VecFilter(PetscVec, PetscReal)
    PetscErrorCode VecReciprocal(PetscVec)
    PetscErrorCode VecPermute(PetscVec, PetscIS, PetscBool)
    PetscErrorCode VecExp(PetscVec)
    PetscErrorCode VecLog(PetscVec)
    PetscErrorCode VecSqrtAbs(PetscVec)
    PetscErrorCode VecAbs(PetscVec)

    PetscErrorCode VecStrideMin(PetscVec, PetscInt, PetscInt*, PetscReal*)
    PetscErrorCode VecStrideMax(PetscVec, PetscInt, PetscInt*, PetscReal*)
    PetscErrorCode VecStrideScale(PetscVec, PetscInt, PetscScalar)
    PetscErrorCode VecStrideGather(PetscVec, PetscInt, PetscVec, PetscInsertMode)
    PetscErrorCode VecStrideScatter(PetscVec, PetscInt, PetscVec, PetscInsertMode)
    PetscErrorCode VecStrideNorm(PetscVec, PetscInt, PetscNormType, PetscReal*)

    PetscErrorCode VecGhostGetLocalForm(PetscVec, PetscVec*)
    PetscErrorCode VecGhostRestoreLocalForm(PetscVec, PetscVec*)
    PetscErrorCode VecGhostUpdateBegin(PetscVec, PetscInsertMode, PetscScatterMode)
    PetscErrorCode VecGhostUpdateEnd(PetscVec, PetscInsertMode, PetscScatterMode)
    PetscErrorCode VecMPISetGhost(PetscVec, PetscInt, const PetscInt*)
    PetscErrorCode VecGhostGetGhostIS(PetscVec, PetscIS*)

    PetscErrorCode VecGetSubVector(PetscVec, PetscIS, PetscVec*)
    PetscErrorCode VecRestoreSubVector(PetscVec, PetscIS, PetscVec*)

    PetscErrorCode VecNestGetSubVecs(PetscVec, PetscInt*, PetscVec**)
    PetscErrorCode VecNestSetSubVecs(PetscVec, PetscInt, PetscInt*, PetscVec*)

    PetscErrorCode VecConcatenate(PetscInt nx, PetscVec[], PetscVec*, PetscIS*[])

    PetscErrorCode VecISAXPY(PetscVec, PetscIS, PetscScalar, PetscVec)
    PetscErrorCode VecISSet(PetscVec, PetscIS, PetscScalar)

    PetscErrorCode VecCUDAGetArrayRead(PetscVec, const PetscScalar*[])
    PetscErrorCode VecCUDAGetArrayWrite(PetscVec, PetscScalar*[])
    PetscErrorCode VecCUDAGetArray(PetscVec, PetscScalar*[])
    PetscErrorCode VecCUDARestoreArrayRead(PetscVec, const PetscScalar*[])
    PetscErrorCode VecCUDARestoreArrayWrite(PetscVec, PetscScalar*[])
    PetscErrorCode VecCUDARestoreArray(PetscVec, PetscScalar*[])

    PetscErrorCode VecHIPGetArrayRead(PetscVec, const PetscScalar*[])
    PetscErrorCode VecHIPGetArrayWrite(PetscVec, PetscScalar*[])
    PetscErrorCode VecHIPGetArray(PetscVec, PetscScalar*[])
    PetscErrorCode VecHIPRestoreArrayRead(PetscVec, const PetscScalar*[])
    PetscErrorCode VecHIPRestoreArrayWrite(PetscVec, PetscScalar*[])
    PetscErrorCode VecHIPRestoreArray(PetscVec, PetscScalar*[])

    PetscErrorCode VecBindToCPU(PetscVec, PetscBool)
    PetscErrorCode VecBoundToCPU(PetscVec, PetscBool*)
    PetscErrorCode VecGetOffloadMask(PetscVec, PetscOffloadMask*)

    PetscErrorCode VecViennaCLGetCLContext(PetscVec, Py_uintptr_t*)
    PetscErrorCode VecViennaCLGetCLQueue(PetscVec, Py_uintptr_t*)
    PetscErrorCode VecViennaCLGetCLMemRead(PetscVec, Py_uintptr_t*)
    PetscErrorCode VecViennaCLGetCLMemWrite(PetscVec, Py_uintptr_t*)
    PetscErrorCode VecViennaCLRestoreCLMemWrite(PetscVec)
    PetscErrorCode VecViennaCLGetCLMem(PetscVec, Py_uintptr_t*)
    PetscErrorCode VecViennaCLRestoreCLMem(PetscVec)

    PetscErrorCode VecCreateSeqCUDAWithArray(MPI_Comm, PetscInt, PetscInt, const PetscScalar*, PetscVec*)
    PetscErrorCode VecCreateMPICUDAWithArray(MPI_Comm, PetscInt, PetscInt, PetscInt, const PetscScalar*, PetscVec*)
    PetscErrorCode VecCreateSeqHIPWithArray(MPI_Comm, PetscInt, PetscInt, const PetscScalar*, PetscVec*)
    PetscErrorCode VecCreateMPIHIPWithArray(MPI_Comm, PetscInt, PetscInt, PetscInt, const PetscScalar*, PetscVec*)

cdef extern from * nogil: # custom.h
    PetscErrorCode VecStrideSum(PetscVec, PetscInt, PetscScalar*)
    PetscErrorCode VecGetCurrentMemType(PetscVec, PetscMemType*)

# --------------------------------------------------------------------

cdef inline Vec ref_Vec(PetscVec vec):
    cdef Vec ob = <Vec> Vec()
    ob.vec = vec
    CHKERR(PetscINCREF(ob.obj))
    return ob

# --------------------------------------------------------------------

# unary operations

cdef Vec vec_pos(Vec self):
    cdef Vec vec = type(self)()
    CHKERR(VecDuplicate(self.vec, &vec.vec))
    CHKERR(VecCopy(self.vec, vec.vec))
    return vec

cdef Vec vec_neg(Vec self):
    cdef Vec vec = <Vec> vec_pos(self)
    CHKERR(VecScale(vec.vec, -1))
    return vec

cdef Vec vec_abs(Vec self):
    cdef Vec vec = <Vec> vec_pos(self)
    CHKERR(VecAbs(vec.vec))
    return vec

# inplace binary operations

cdef Vec vec_iadd(Vec self, other):
    cdef PetscScalar alpha = 1
    cdef Vec vec
    if isinstance(other, Vec):
        alpha = 1; vec = other
        CHKERR(VecAXPY(self.vec, alpha, vec.vec))
    elif isinstance(other, (tuple, list)):
        other, vec = other
        alpha = asScalar(other)
        CHKERR(VecAXPY(self.vec, alpha, vec.vec))
    else:
        alpha = asScalar(other)
        CHKERR(VecShift(self.vec, alpha))
    return self

cdef Vec vec_isub(Vec self, other):
    cdef PetscScalar alpha = 1
    cdef Vec vec
    if isinstance(other, Vec):
        alpha = 1; vec = other
        CHKERR(VecAXPY(self.vec, -alpha, vec.vec))
    elif isinstance(other, (tuple, list)):
        other, vec = other
        alpha = asScalar(other)
        CHKERR(VecAXPY(self.vec, -alpha, vec.vec))
    else:
        alpha = asScalar(other)
        CHKERR(VecShift(self.vec, -alpha))
    return self

cdef Vec vec_imul(Vec self, other):
    cdef PetscScalar alpha = 1
    cdef Vec vec
    if isinstance(other, Vec):
        vec = other
        CHKERR(VecPointwiseMult(self.vec, self.vec, vec.vec))
    else:
        alpha = asScalar(other)
        CHKERR(VecScale(self.vec, alpha))
    return self

cdef Vec vec_idiv(Vec self, other):
    cdef PetscScalar one = 1
    cdef PetscScalar alpha = 1
    cdef Vec vec
    if isinstance(other, Vec):
        vec = other
        CHKERR(VecPointwiseDivide(self.vec, self.vec, vec.vec))
    else:
        alpha = asScalar(other)
        CHKERR(VecScale(self.vec, one/alpha))
    return self

# binary operations

cdef Vec vec_add(Vec self, other):
    return vec_iadd(vec_pos(self), other)

cdef Vec vec_sub(Vec self, other):
    return vec_isub(vec_pos(self), other)

cdef Vec vec_mul(Vec self, other):
    return vec_imul(vec_pos(self), other)

cdef Vec vec_div(Vec self, other):
    return vec_idiv(vec_pos(self), other)

cdef object vec_matmul(Vec self, other):
    if isinstance(other, Vec):
        return self.dot(other)
    if isinstance(other, Mat):
        result = other.createVecRight()
        other.multTranspose(self, result)
        return result
    return NotImplemented

# reflected binary operations

cdef Vec vec_radd(Vec self, other):
    return vec_add(self, other)

cdef Vec vec_rsub(Vec self, other):
    cdef Vec vec = <Vec> vec_sub(self, other)
    CHKERR(VecScale(vec.vec, -1))
    return vec

cdef Vec vec_rmul(Vec self, other):
    return vec_mul(self, other)

cdef Vec vec_rdiv(Vec self, other):
    cdef Vec vec = <Vec> vec_div(self, other)
    CHKERR(VecReciprocal(vec.vec))
    return vec

# --------------------------------------------------------------------

cdef inline int Vec_Sizes(object size, object bsize,
                          PetscInt *b, PetscInt *n, PetscInt *N) except -1:
    Sys_Sizes(size, bsize, b, n, N)
    return 0

# --------------------------------------------------------------------

ctypedef PetscErrorCode VecSetValuesFcn(PetscVec,
                                        PetscInt, const PetscInt*,
                                        const PetscScalar*, PetscInsertMode)

cdef inline VecSetValuesFcn* vecsetvalues_fcn(int blocked, int local):
    cdef VecSetValuesFcn *setvalues = NULL
    if blocked and local: setvalues = VecSetValuesBlockedLocal
    elif blocked:         setvalues = VecSetValuesBlocked
    elif local:           setvalues = VecSetValuesLocal
    else:                 setvalues = VecSetValues
    return setvalues

cdef inline int vecsetvalues(PetscVec V,
                             object oi, object ov, object oim,
                             int blocked, int local) except -1:
    # block size
    cdef PetscInt bs=1
    if blocked:
        CHKERR(VecGetBlockSize(V, &bs))
        if bs < 1: bs = 1
    # indices and values
    cdef PetscInt ni=0, nv=0
    cdef PetscInt *i=NULL
    cdef PetscScalar *v=NULL
    cdef object unused1 = iarray_i(oi, &ni, &i)
    cdef object unused2 = iarray_s(ov, &nv, &v)
    if ni*bs != nv: raise ValueError(
        "incompatible array sizes: ni=%d, nv=%d, bs=%d" %
        (toInt(ni), toInt(nv), toInt(bs)))
    # VecSetValuesXXX function and insert mode
    cdef VecSetValuesFcn *setvalues = vecsetvalues_fcn(blocked, local)
    cdef PetscInsertMode addv = insertmode(oim)
    # actual call
    CHKERR(setvalues(V, ni, i, v, addv))
    return 0

cdef object vecgetvalues(PetscVec vec, object oindices, object values):
    cdef PetscInt ni=0, nv=0
    cdef PetscInt *i=NULL
    cdef PetscScalar *v=NULL
    cdef object indices = iarray_i(oindices, &ni, &i)
    if values is None:
        values = empty_s(ni)
        values.shape = indices.shape
    values = oarray_s(values, &nv, &v)
    if (ni != nv): raise ValueError(
        ("incompatible array sizes: "
         "ni=%d, nv=%d") % (toInt(ni), toInt(nv)))
    CHKERR(VecGetValues(vec, ni, i, v))
    return values

# --------------------------------------------------------------------

cdef inline _Vec_buffer vec_getbuffer_r(Vec self):
    cdef _Vec_buffer buf = _Vec_buffer(self)
    buf.readonly = 1
    return buf

cdef inline _Vec_buffer vec_getbuffer_w(Vec self):
    cdef _Vec_buffer buf = _Vec_buffer(self)
    buf.readonly = 0
    return buf

cdef inline ndarray vec_getarray_r(Vec self):
    return asarray(vec_getbuffer_r(self))

cdef inline ndarray vec_getarray_w(Vec self):
    return asarray(vec_getbuffer_w(self))

cdef inline int vec_setarray(Vec self, object o) except -1:
    cdef PetscInt na=0, nv=0, i=0
    cdef PetscScalar *va=NULL, *vv=NULL
    cdef ndarray ary = iarray_s(o, &na, &va)
    CHKERR(VecGetLocalSize(self.vec, &nv))
    if (na != nv) and PyArray_NDIM(ary) > 0: raise ValueError(
        "array size %d incompatible with vector local size %d" %
        (toInt(na), toInt(nv)))
    CHKERR(VecGetArray(self.vec, &vv))
    try:
        if PyArray_NDIM(ary) == 0:
            for i from 0 <= i < nv:
                vv[i] = va[0]
        else:
            CHKERR(PetscMemcpy(vv, va, <size_t>nv*sizeof(PetscScalar)))
    finally:
        CHKERR(VecRestoreArray(self.vec, &vv))
    return 0

cdef object vec_getitem(Vec self, object i):
    cdef PetscInt N=0
    if i is Ellipsis:
        return asarray(self)
    if isinstance(i, slice):
        CHKERR(VecGetSize(self.vec, &N))
        start, stop, stride = i.indices(toInt(N))
        i = arange(start, stop, stride)
    return vecgetvalues(self.vec, i, None)

cdef int vec_setitem(Vec self, object i, object v) except -1:
    cdef PetscInt N=0
    if i is Ellipsis:
        return vec_setarray(self, v)
    if isinstance(i, slice):
        CHKERR(VecGetSize(self.vec, &N))
        start, stop, stride = i.indices(toInt(N))
        i = arange(start, stop, stride)
    vecsetvalues(self.vec, i, v, None, 0, 0)
    return 0

cdef vec_get_dlpack_ctx(Vec self):
    cdef object ctx0 = self.get_attr('__dltensor_ctx__')
    cdef PetscInt n = 0
    cdef int64_t ndim = 1
    cdef int64_t* shape_arr = NULL
    cdef int64_t* strides_arr = NULL
    cdef object s1 = None
    cdef object s2 = None
    cdef PetscInt devId = 0
    cdef PetscMemType mtype = PETSC_MEMTYPE_HOST
    if ctx0 is None: # First time in, create a linear memory view
        s1 = oarray_p(empty_p(<PetscInt>ndim), NULL, <void**>&shape_arr)
        s2 = oarray_p(empty_p(<PetscInt>ndim), NULL, <void**>&strides_arr)
        CHKERR(VecGetLocalSize(self.vec, &n))
        shape_arr[0] = <int64_t>n
        strides_arr[0] = 1
    else:
        (_, _, ndim, s1, s2) = ctx0

    devType_ = {PETSC_MEMTYPE_HOST : kDLCPU, PETSC_MEMTYPE_CUDA : kDLCUDA, PETSC_MEMTYPE_HIP : kDLROCM}
    CHKERR(VecGetCurrentMemType(self.vec, &mtype))
    dtype = devType_.get(mtype, kDLCPU)
    if dtype != kDLCPU:
        CHKERR(PetscObjectGetDeviceId(<PetscObject>self.vec, &devId))
    ctx0 = (dtype, devId, ndim, s1, s2)
    self.set_attr('__dltensor_ctx__', ctx0)
    return ctx0

# --------------------------------------------------------------------

cdef int Vec_AcquireArray(PetscVec v, PetscScalar *a[], int ro) except -1 nogil:
    if ro: CHKERR(VecGetArrayRead(v, <const PetscScalar**>a))
    else:  CHKERR(VecGetArray(v, a))
    return 0

cdef int Vec_ReleaseArray(PetscVec v, PetscScalar *a[], int ro) except -1 nogil:
    if ro: CHKERR(VecRestoreArrayRead(v, <const PetscScalar**>a))
    else:  CHKERR(VecRestoreArray(v, a))
    return 0

cdef class _Vec_buffer:

    cdef PetscVec vec
    cdef PetscInt size
    cdef PetscScalar *data
    cdef bint readonly
    cdef bint hasarray

    def __cinit__(self, Vec vec, bint readonly=0):
        cdef PetscVec v = vec.vec
        CHKERR(PetscINCREF(<PetscObject*>&v))
        self.vec = v
        self.size = 0
        self.data = NULL
        self.readonly = 1 if readonly else 0
        self.hasarray = 0

    def __dealloc__(self):
        if self.hasarray and self.vec != NULL:
            Vec_ReleaseArray(self.vec, &self.data, self.readonly)
        CHKERR(VecDestroy(&self.vec))

    #

    cdef int acquire(self) except -1 nogil:
        if not self.hasarray and self.vec != NULL:
            CHKERR(VecGetLocalSize(self.vec, &self.size))
            Vec_AcquireArray(self.vec, &self.data, self.readonly)
            self.hasarray = 1
        return 0

    cdef int release(self) except -1 nogil:
        if self.hasarray and self.vec != NULL:
            self.size = 0
            Vec_ReleaseArray(self.vec, &self.data, self.readonly)
            self.hasarray = 0
        return 0

    # buffer interface (PEP 3118)

    cdef int acquirebuffer(self, Py_buffer *view, int flags) except -1:
        self.acquire()
        PyPetscBuffer_FillInfo(view, <void*>self.data, self.size,
                               c's', self.readonly, flags)
        view.obj = self
        return 0

    cdef int releasebuffer(self, Py_buffer *view) except -1:
        PyPetscBuffer_Release(view)
        self.release()
        return 0

    def __getbuffer__(self, Py_buffer *view, int flags):
        self.acquirebuffer(view, flags)

    def __releasebuffer__(self, Py_buffer *view):
        self.releasebuffer(view)

    # 'with' statement (PEP 343)

    cdef object enter(self):
        self.acquire()
        return asarray(self)

    cdef object exit(self):
        self.release()
        return None

    def __enter__(self):
        return self.enter()

    def __exit__(self, *exc):
        return self.exit()

    # NumPy array interface (legacy)

    property __array_interface__:
        def __get__(self):
            cdef PetscInt n = 0
            if self.vec != NULL:
                CHKERR(VecGetLocalSize(self.vec, &n))
            cdef object size = toInt(n)
            cdef dtype descr = PyArray_DescrFromType(NPY_PETSC_SCALAR)
            cdef str typestr = "=%c%d" % (descr.kind, descr.itemsize)
            return dict(version=3,
                        data=self,
                        shape=(size,),
                        typestr=typestr)

# --------------------------------------------------------------------

cdef class _Vec_LocalForm:

    cdef Vec gvec
    cdef Vec lvec

    def __init__(self, Vec gvec):
        self.gvec = gvec
        self.lvec = Vec()

    def __enter__(self):
        cdef PetscVec gvec = self.gvec.vec
        CHKERR(VecGhostGetLocalForm(gvec, &self.lvec.vec))
        return self.lvec

    def __exit__(self, *exc):
        cdef PetscVec gvec = self.gvec.vec
        CHKERR(VecGhostRestoreLocalForm(gvec, &self.lvec.vec))
        self.lvec.vec = NULL



# --------------------------------------------------------------------

class DMDAStencilType(object):
    """Stencil types."""
    STAR = DMDA_STENCIL_STAR
    BOX  = DMDA_STENCIL_BOX


class DMDAInterpolationType(object):
    """Interpolation types."""
    Q0 = DMDA_INTERPOLATION_Q0
    Q1 = DMDA_INTERPOLATION_Q1


class DMDAElementType(object):
    """Element types."""
    P1 = DMDA_ELEMENT_P1
    Q1 = DMDA_ELEMENT_Q1

# --------------------------------------------------------------------


cdef class DMDA(DM):
    """A DM object that is used to manage data for a structured grid."""

    StencilType       = DMDAStencilType
    InterpolationType = DMDAInterpolationType
    ElementType       = DMDAElementType

    #

    def create(
        self,
        dim: int | None = None,
        dof: int | None = None,
        sizes: DimsSpec | None = None,
        proc_sizes: DimsSpec | None = None,
        boundary_type: tuple[DM.BoundaryType | int | str | bool, ...] | None = None,
        stencil_type: StencilType | None = None,
        stencil_width: int | None = None,
        bint setup: bool = True,
        ownership_ranges: tuple[Sequence[int], ...] | None = None,
        comm: Comm | None = None) -> Self:
        """Create a ``DMDA`` object.

        Collective.

        This routine performs the following steps of the C API:
        - ``petsc.DMDACreate``
        - ``petsc.DMSetDimension``
        - ``petsc.DMDASetDof``
        - ``petsc.DMDASetSizes``
        - ``petsc.DMDASetNumProcs``
        - ``petsc.DMDASetOwnershipRanges``
        - ``petsc.DMDASetBoundaryType``
        - ``petsc.DMDASetStencilType``
        - ``petsc.DMDASetStencilWidth``
        - ``petsc.DMSetUp`` (optionally)

        Parameters
        ----------
        dim
            The number of dimensions.
        dof
            The number of degrees of freedom.
        sizes
            The number of elements in each dimension.
        proc_sizes
            The number of processes in x, y, z dimensions.
        boundary_type
            The boundary types.
        stencil_type
            The ghost/halo stencil type.
        stencil_width
            The width of the ghost/halo region.
        setup
            Whether to call the setup routine after creating the object.
        ownership_ranges
            Local x, y, z element counts, of length equal to ``proc_sizes``,
            summing to ``sizes``.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.DMDACreate, petsc.DMSetDimension, petsc.DMDASetDof
        petsc.DMDASetSizes, petsc.DMDASetNumProcs
        petsc.DMDASetOwnershipRanges, petsc.DMDASetBoundaryType
        petsc.DMDASetStencilType, petsc.DMDASetStencilWidth, petsc.DMSetUp

        """
        #
        cdef object arg = None
        try: arg = tuple(dim)
        except TypeError: pass
        else: dim, sizes = None, arg
        #
        cdef PetscInt ndim = PETSC_DECIDE
        cdef PetscInt ndof = PETSC_DECIDE
        cdef PetscInt M = 1, m = PETSC_DECIDE, *lx = NULL
        cdef PetscInt N = 1, n = PETSC_DECIDE, *ly = NULL
        cdef PetscInt P = 1, p = PETSC_DECIDE, *lz = NULL
        cdef PetscDMBoundaryType btx = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType bty = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType btz = DM_BOUNDARY_NONE
        cdef PetscDMDAStencilType stype  = DMDA_STENCIL_BOX
        cdef PetscInt             swidth = PETSC_DECIDE
        # grid and proc sizes
        cdef object gsizes = sizes
        cdef object psizes = proc_sizes
        cdef PetscInt gdim = PETSC_DECIDE
        cdef PetscInt pdim = PETSC_DECIDE
        if sizes is not None:
            gdim = asDims(gsizes, &M, &N, &P)
        if psizes is not None:
            pdim = asDims(psizes, &m, &n, &p)
        if gdim>=0 and pdim>=0:
            assert gdim == pdim
        # dim and dof
        if dim is not None: ndim = asInt(dim)
        if dof is not None: ndof = asInt(dof)
        if ndim==PETSC_DECIDE: ndim = gdim
        if ndof==PETSC_DECIDE: ndof = 1
        # vertex distribution
        if ownership_ranges is not None:
            ownership_ranges = asOwnershipRanges(ownership_ranges,
                                                 ndim, &m, &n, &p,
                                                 &lx, &ly, &lz)
        # periodicity, stencil type & width
        if boundary_type is not None:
            asBoundary(boundary_type, &btx, &bty, &btz)
        if stencil_type is not None:
            stype = asStencil(stencil_type)
        if stencil_width is not None:
            swidth = asInt(stencil_width)
        if setup and swidth == PETSC_DECIDE: swidth = 0
        # create the DMDA object
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscDM newda = NULL
        CHKERR(DMDACreateND(ccomm, ndim, ndof,
                            M, N, P, m, n, p, lx, ly, lz,
                            btx, bty, btz, stype, swidth,
                            &newda))
        if setup and ndim > 0: CHKERR(DMSetUp(newda))
        CHKERR(PetscCLEAR(self.obj)); self.dm = newda
        return self

    def duplicate(
        self,
        dof: int | None = None,
        boundary_type: tuple[DM.BoundaryType | int | str | bool, ...] | None = None,
        stencil_type: StencilType | None = None,
        stencil_width: int | None = None) -> DMDA:
        """Duplicate a DMDA.

        Collective.

        This routine retrieves the information from the DMDA and recreates it.
        Parameters ``dof``, ``boundary_type``, ``stencil_type``,
        ``stencil_width`` will be overwritten, if provided.

        Parameters
        ----------
        dof
            The number of degrees of freedom.
        boundary_type
            Boundary types.
        stencil_type
            The ghost/halo stencil type.
        stencil_width
            The width of the ghost/halo region.

        See Also
        --------
        create, petsc.DMDAGetInfo, petsc.DMSetUp

        """
        cdef PetscInt ndim = 0, ndof = 0
        cdef PetscInt M = 1, N = 1, P = 1
        cdef PetscInt m = 1, n = 1, p = 1
        cdef PetscDMBoundaryType btx = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType bty = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType btz = DM_BOUNDARY_NONE
        cdef PetscDMDAStencilType  stype  = DMDA_STENCIL_BOX
        cdef PetscInt              swidth = PETSC_DECIDE
        CHKERR(DMDAGetInfo(self.dm,
                           &ndim,
                           &M, &N, &P,
                           &m, &n, &p,
                           &ndof, &swidth,
                           &btx, &bty, &btz,
                           &stype))
        cdef const PetscInt *lx = NULL, *ly = NULL, *lz = NULL
        CHKERR(DMDAGetOwnershipRanges(self.dm, &lx, &ly, &lz))
        cdef MPI_Comm comm = MPI_COMM_NULL
        CHKERR(PetscObjectGetComm(<PetscObject>self.dm, &comm))
        #
        if dof is not None:
            ndof = asInt(dof)
        if boundary_type is not None:
            asBoundary(boundary_type, &btx, &bty, &btz)
        if stencil_type  is not None:
            stype = asStencil(stencil_type)
        if stencil_width is not None:
            swidth = asInt(stencil_width)
        #
        cdef DMDA da = DMDA()
        CHKERR(DMDACreateND(comm, ndim, ndof,
                            M, N, P, m, n, p, lx, ly, lz,
                            btx, bty, btz, stype, swidth,
                            &da.dm))
        CHKERR(DMSetUp(da.dm))
        return da

    #

    def setDim(self, dim: int) -> None:
        """Set the topological dimension.

        Collective.

        Parameters
        ----------
        dim
            Topological dimension.

        See Also
        --------
        getDim, petsc.DMSetDimension

        """
        return self.setDimension(dim)

    def getDim(self) -> int:
        """Return the topological dimension.

        Not collective.

        See Also
        --------
        setDim, petsc.DMGetDimension

        """
        return self.getDimension()

    def setDof(self, dof: int) -> None:
        """Set the number of degrees of freedom per vertex.

        Not collective.

        Parameters
        ----------
        dof
            The number of degrees of freedom.

        See Also
        --------
        getDof, petsc.DMDASetDof

        """
        cdef PetscInt ndof = asInt(dof)
        CHKERR(DMDASetDof(self.dm, ndof))

    def getDof(self) -> int:
        """Return the number of degrees of freedom per node.

        Not collective.

        See Also
        --------
        setDof, petsc.DMDAGetInfo

        """
        cdef PetscInt dof = 0
        CHKERR(DMDAGetInfo(self.dm,
                           NULL,
                           NULL, NULL, NULL,
                           NULL, NULL, NULL,
                           &dof, NULL,
                           NULL, NULL, NULL,
                           NULL))
        return toInt(dof)

    def setSizes(self, sizes: DimsSpec) -> None:
        """Set the number of grid points in each dimension.

        Logically collective.

        Parameters
        ----------
        sizes
            The global (x,), (x, y), or (x, y, z) size.

        See Also
        --------
        getSizes, petsc.DMDASetSizes

        """
        cdef tuple gsizes = tuple(sizes)
        cdef PetscInt gdim = PETSC_DECIDE
        cdef PetscInt M = 1
        cdef PetscInt N = 1
        cdef PetscInt P = 1
        gdim = asDims(gsizes, &M, &N, &P)
        cdef PetscInt dim = PETSC_DECIDE
        CHKERR(DMDAGetDim(self.dm, &dim))
        if dim == PETSC_DECIDE:
            CHKERR(DMSetDimension(self.dm, gdim))
        CHKERR(DMDASetSizes(self.dm, M, N, P))

    def getSizes(self) -> tuple[int, ...]:
        """Return the number of grid points in each dimension.

        Not collective.

        See Also
        --------
        setSizes, petsc.DMDAGetInfo

        """
        cdef PetscInt dim = 0
        cdef PetscInt M = PETSC_DECIDE
        cdef PetscInt N = PETSC_DECIDE
        cdef PetscInt P = PETSC_DECIDE
        CHKERR(DMDAGetInfo(self.dm,
                           &dim,
                           &M, &N, &P,
                           NULL, NULL, NULL,
                           NULL, NULL,
                           NULL, NULL, NULL,
                           NULL))
        return toDims(dim, M, N, P)

    def setProcSizes(self, proc_sizes: DimsSpec) -> None:
        """Set the number of processes in each dimension.

        Logically collective.

        Parameters
        ----------
        proc_sizes
            The number of processes in (x,), (x, y), or (x, y, z) dimensions.

        See Also
        --------
        getProcSizes, petsc.DMDASetNumProcs

        """
        cdef tuple psizes = tuple(proc_sizes)
        cdef PetscInt pdim = PETSC_DECIDE
        cdef PetscInt m = PETSC_DECIDE
        cdef PetscInt n = PETSC_DECIDE
        cdef PetscInt p = PETSC_DECIDE
        pdim = asDims(psizes, &m, &n, &p)
        cdef PetscInt dim = PETSC_DECIDE
        CHKERR(DMDAGetDim(self.dm, &dim))
        if dim == PETSC_DECIDE:
            CHKERR(DMSetDimension(self.dm, pdim))
        CHKERR(DMDASetNumProcs(self.dm, m, n, p))

    def getProcSizes(self) -> tuple[int, ...]:
        """Return the number of processes in each dimension.

        Not collective.

        See Also
        --------
        setProcSizes, petsc.DMDAGetInfo

        """
        cdef PetscInt dim = 0
        cdef PetscInt m = PETSC_DECIDE
        cdef PetscInt n = PETSC_DECIDE
        cdef PetscInt p = PETSC_DECIDE
        CHKERR(DMDAGetInfo(self.dm,
                           &dim,
                           NULL, NULL, NULL,
                           &m, &n, &p,
                           NULL, NULL,
                           NULL, NULL, NULL,
                           NULL))
        return toDims(dim, m, n, p)

    def setBoundaryType(
        self,
        boundary_type: tuple[DM.BoundaryType | int | str | bool, ...]) -> None:
        """Set the type of ghost nodes on domain boundaries.

        Not collective.

        Parameters
        ----------
        boundary_type
            The boundary type in (x), (x, y), or (x, y, z) dimensions.

        See Also
        --------
        getBoundaryType, petsc.DMDASetBoundaryType

        """
        cdef PetscDMBoundaryType btx = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType bty = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType btz = DM_BOUNDARY_NONE
        asBoundary(boundary_type, &btx, &bty, &btz)
        CHKERR(DMDASetBoundaryType(self.dm, btx, bty, btz))

    def getBoundaryType(self) -> tuple[DM.BoundaryType, ...]:
        """Return the type of ghost nodes at boundary in each dimension.

        Not collective.

        See Also
        --------
        setBoundaryType

        """
        cdef PetscInt dim = 0
        cdef PetscDMBoundaryType btx = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType bty = DM_BOUNDARY_NONE
        cdef PetscDMBoundaryType btz = DM_BOUNDARY_NONE
        CHKERR(DMDAGetInfo(self.dm,
                           &dim,
                           NULL, NULL, NULL,
                           NULL, NULL, NULL,
                           NULL, NULL,
                           &btx, &bty, &btz,
                           NULL))
        return toDims(dim, btx, bty, btz)

    def setStencilType(self, stencil_type: StencilType) -> None:
        """Set the stencil type.

        Logically collective.

        Parameters
        ----------
        stype
            The stencil type.

        See Also
        --------
        getStencilType, setStencil, petsc.DMDASetStencilType

        """
        cdef PetscDMDAStencilType stype = asStencil(stencil_type)
        CHKERR(DMDASetStencilType(self.dm, stype))

    def getStencilType(self) -> StencilType:
        """Return the stencil type.

        Not collective.

        See Also
        --------
        setStencilType, petsc.DMDAGetInfo

        """
        cdef PetscDMDAStencilType stype = DMDA_STENCIL_BOX
        CHKERR(DMDAGetInfo(self.dm,
                           NULL,
                           NULL, NULL, NULL,
                           NULL, NULL, NULL,
                           NULL, NULL,
                           NULL, NULL, NULL,
                           &stype))
        return stype

    def setStencilWidth(self, stencil_width: int) -> None:
        """Set the stencil width.

        Logically collective.

        Parameters
        ----------
        stencil_width
            The stencil width.

        See Also
        --------
        getStencilWidth, setStencil, petsc.DMDASetStencilWidth

        """
        cdef PetscInt swidth = asInt(stencil_width)
        CHKERR(DMDASetStencilWidth(self.dm, swidth))

    def getStencilWidth(self) -> int:
        """Return the stencil width.

        Not collective.

        See Also
        --------
        setStencilWidth

        """
        cdef PetscInt swidth = 0
        CHKERR(DMDAGetInfo(self.dm,
                           NULL,
                           NULL, NULL, NULL,
                           NULL, NULL, NULL,
                           NULL, &swidth,
                           NULL, NULL, NULL,
                           NULL))
        return toInt(swidth)

    def setStencil(
        self,
        stencil_type: StencilType,
        stencil_width: int) -> None:
        """Set the stencil type and width.

        Not collective.

        Parameters
        ----------
        stencil_type
            The stencil type.
        stencil_width
            The stencil width.

        See Also
        --------
        setStencilWidth, setStencilType, petsc.DMDASetStencilType
        petsc.DMDASetStencilWidth

        """
        cdef PetscDMDAStencilType stype = asStencil(stencil_type)
        cdef PetscInt swidth = asInt(stencil_width)
        CHKERR(DMDASetStencilType(self.dm, stype))
        CHKERR(DMDASetStencilWidth(self.dm, swidth))

    def getStencil(self) -> tuple[StencilType, int]:
        """Return the stencil type and width.

        Not collective.

        See Also
        --------
        getStencilType, getStencilWidth

        """
        cdef PetscDMDAStencilType stype = DMDA_STENCIL_BOX
        cdef PetscInt swidth = 0
        CHKERR(DMDAGetInfo(self.dm,
                           NULL,
                           NULL, NULL, NULL,
                           NULL, NULL, NULL,
                           NULL, &swidth,
                           NULL, NULL, NULL,
                           &stype))
        return (toStencil(stype), toInt(swidth))

    #

    def getRanges(self) -> tuple[tuple[int, int], ...]:
        """Return the ranges of the owned local region in each dimension.

        Not collective.

        Excluding ghost nodes.

        See Also
        --------
        getGhostRanges, getOwnershipRanges, getCorners, getGhostCorners
        petsc.DMDAGetCorners

        """
        cdef PetscInt dim=0, x=0, y=0, z=0, m=0, n=0, p=0
        CHKERR(DMDAGetDim(self.dm, &dim))
        CHKERR(DMDAGetCorners(self.dm,
                              &x, &y, &z,
                              &m, &n, &p))
        return ((toInt(x), toInt(x+m)),
                (toInt(y), toInt(y+n)),
                (toInt(z), toInt(z+p)))[:<Py_ssize_t>dim]

    def getGhostRanges(self) -> tuple[tuple[int, int], ...]:
        """Return the ranges of the local region in each dimension, including ghost nodes.

        Not collective.

        See Also
        --------
        getRanges, getOwnershipRanges, getCorners, getGhostCorners
        petsc.DMDAGetGhostCorners

        """
        cdef PetscInt dim=0, x=0, y=0, z=0, m=0, n=0, p=0
        CHKERR(DMDAGetDim(self.dm, &dim))
        CHKERR(DMDAGetGhostCorners(self.dm,
                                   &x, &y, &z,
                                   &m, &n, &p))
        return ((toInt(x), toInt(x+m)),
                (toInt(y), toInt(y+n)),
                (toInt(z), toInt(z+p)))[:<Py_ssize_t>dim]

    def getOwnershipRanges(self) -> tuple[ArrayInt, ...]:
        """Return the ranges of indices in each dimension owned by each process.

        Not collective.

        These numbers are not multiplied by the number of DOFs per node.

        See Also
        --------
        getRanges, getGhostRanges, getCorners, getGhostCorners
        petsc.DMDAGetOwnershipRanges

        """
        cdef PetscInt dim=0, m=0, n=0, p=0
        cdef const PetscInt *lx = NULL, *ly = NULL, *lz = NULL
        CHKERR(DMDAGetInfo(self.dm,
                           &dim,
                           NULL, NULL, NULL,
                           &m, &n, &p,
                           NULL, NULL,
                           NULL, NULL, NULL,
                           NULL))
        CHKERR(DMDAGetOwnershipRanges(self.dm, &lx, &ly, &lz))
        return toOwnershipRanges(dim, m, n, p, lx, ly, lz)

    def getCorners(self) -> tuple[tuple[int, ...], tuple[int, ...]]:
        """Return the lower left corner and the sizes of the owned local region.

        Not collective.

        Returns the global (x,y,z) indices of the lower left corner (first
        tuple) and size of the local region (second tuple).

        Excluding ghost points.

        The corner information is independent of the number of degrees of
        freedom per node. Thus the returned values can be thought of as
        coordinates on a logical grid, where each grid point has (potentially)
        several degrees of freedom.

        See Also
        --------
        getRanges, getGhostRanges, getOwnershipRanges, getGhostCorners
        petsc.DMDAGetCorners

        """
        cdef PetscInt dim=0, x=0, y=0, z=0, m=0, n=0, p=0
        CHKERR(DMDAGetDim(self.dm, &dim))
        CHKERR(DMDAGetCorners(self.dm,
                              &x, &y, &z,
                              &m, &n, &p))
        return ((toInt(x), toInt(y), toInt(z))[:<Py_ssize_t>dim],
                (toInt(m), toInt(n), toInt(p))[:<Py_ssize_t>dim])

    def getGhostCorners(self) -> tuple[tuple[int, ...], tuple[int, ...]]:
        """Return the lower left corner and the size of the ghosted local region.

        Not collective.

        Returns the global (x,y,z) indices of the lower left corner (first
        tuple) and size of the local region (second tuple).

        See Also
        --------
        getRanges, getGhostRanges, getOwnershipRanges, getCorners
        petsc.DMDAGetGhostCorners

        """
        cdef PetscInt dim=0, x=0, y=0, z=0, m=0, n=0, p=0
        CHKERR(DMDAGetDim(self.dm, &dim))
        CHKERR(DMDAGetGhostCorners(self.dm,
                                   &x, &y, &z,
                                   &m, &n, &p))
        return ((toInt(x), toInt(y), toInt(z))[:<Py_ssize_t>dim],
                (toInt(m), toInt(n), toInt(p))[:<Py_ssize_t>dim])

    #

    def setFieldName(self, field: int, name: str) -> None:
        """Set the name of individual field components.

        Logically collective.

        Parameters
        ----------
        field
            The field number for the DMDA (``0``, ``1``, ..., ``dof-1``),
            where ``dof`` indicates the number of degrees of freedom per node
            within the `DMDA`.
        name
            The name of the field (component).

        See Also
        --------
        getFieldName, petsc.DMDASetFieldName

        """
        cdef PetscInt ival = asInt(field)
        cdef const char *cval = NULL
        name = str2bytes(name, &cval)
        CHKERR(DMDASetFieldName(self.dm, ival, cval))

    def getFieldName(self, field: int) -> str:
        """Return the name of an individual field component.

        Not collective.

        Parameters
        ----------
        field
            The field number for the DMDA (``0``, ``1``, ..., ``dof-1``),
            where ``dof`` indicates the number of degrees of freedom per node
            within the `DMDA`.

        See Also
        --------
        setFieldName, petsc.DMDAGetFieldName

        """
        cdef PetscInt ival = asInt(field)
        cdef const char *cval = NULL
        CHKERR(DMDAGetFieldName(self.dm, ival, &cval))
        return bytes2str(cval)

    #

    def getVecArray(self, Vec vec, readonly: bool = False) -> Any:
        """Get access to the vector as laid out on a N-d grid.

        Logically collective.

        Parameters
        ----------
        vec
            The vector to which access is being requested.
        readonly
            Request read-only access.

        See Also
        --------
        Vec.getArray, petsc.DMDAVecGetArray, petsc.DMDAVecGetArrayDOF

        """
        return _DMDA_Vec_array(self, vec, readonly)

    #

    def setUniformCoordinates(
        self,
        xmin: float = 0,
        xmax: float = 1,
        ymin: float = 0,
        ymax: float = 1,
        zmin: float = 0,
        zmax: float = 1) -> None:
        """Set the DMDA coordinates to be a uniform grid.

        Collective.

        Parameters
        ----------
        xmin
            The minimum in the ``x`` dimension.
        xmax
            The maximum in the ``x`` dimension.
        ymin
            The minimum in the ``y`` dimension (value ignored for 1 dimensional
            problems).
        ymax
            The maximum in the ``y`` dimension (value ignored for 1 dimensional
            problems).
        zmin
            The minimum in the ``z`` dimension (value ignored for 1 or 2
            dimensional problems).
        zmax
            The maximum in the ``z`` dimension (value ignored for 1 or 2
            dimensional problems).

        See Also
        --------
        petsc.DMDASetUniformCoordinates

        """
        cdef PetscReal _xmin = asReal(xmin), _xmax = asReal(xmax)
        cdef PetscReal _ymin = asReal(ymin), _ymax = asReal(ymax)
        cdef PetscReal _zmin = asReal(zmin), _zmax = asReal(zmax)
        CHKERR(DMDASetUniformCoordinates(self.dm,
                                         _xmin, _xmax,
                                         _ymin, _ymax,
                                         _zmin, _zmax))

    def setCoordinateName(self, index: int, name: str) -> None:
        """Set the name of the coordinate dimension.

        Logically collective.

        Parameters
        ----------
        index
            The coordinate number for the DMDA (``0``, ``1``, ..., ``dim-1``).
        name
            The name of the coordinate.

        See Also
        --------
        getCoordinateName, petsc.DMDASetCoordinateName

        """
        cdef PetscInt ival = asInt(index)
        cdef const char *cval = NULL
        name = str2bytes(name, &cval)
        CHKERR(DMDASetCoordinateName(self.dm, ival, cval))

    def getCoordinateName(self, index: int) -> str:
        """Return the name of a coordinate dimension.

        Not collective.

        Parameters
        ----------
        index
            The coordinate number for the DMDA (``0``, ``1``, ..., ``dim-1``).

        See Also
        --------
        setCoordinateName, petsc.DMDAGetCoordinateName

        """
        cdef PetscInt ival = asInt(index)
        cdef const char *cval = NULL
        CHKERR(DMDAGetCoordinateName(self.dm, ival, &cval))
        return bytes2str(cval)

    #

    def createNaturalVec(self) -> Vec:
        """Create a vector that will hold values in the natural numbering.

        Collective.

        The number of local entries in the vector on each process is the same
        as in a vector created with `DM.createGlobalVec`.

        See Also
        --------
        petsc.DMDACreateNaturalVector

        """
        cdef Vec vn = Vec()
        CHKERR(DMDACreateNaturalVector(self.dm, &vn.vec))
        return vn

    def globalToNatural(
        self,
        Vec vg,
        Vec vn,
        addv: InsertMode | None = None) -> None:
        """Map values to the "natural" grid ordering.

        Neighborwise collective.

        You must call `createNaturalVec` before using this routine.

        Parameters
        ----------
        vg
            The global vector in a grid ordering.
        vn
            The global vector in a "natural" ordering.
        addv
            The insertion mode.

        See Also
        --------
        naturalToGlobal, petsc.DMDAGlobalToNaturalBegin
        petsc.DMDAGlobalToNaturalEnd

        """
        cdef PetscInsertMode im = insertmode(addv)
        CHKERR(DMDAGlobalToNaturalBegin(self.dm, vg.vec, im, vn.vec))
        CHKERR(DMDAGlobalToNaturalEnd(self.dm, vg.vec, im, vn.vec))

    def naturalToGlobal(
        self,
        Vec vn,
        Vec vg,
        addv: InsertMode | None = None) -> None:
        """Map values the to grid ordering.

        Neighborwise collective.

        Parameters
        ----------
        vn
            The global vector in a natural ordering.
        vg
            the global vector in a grid ordering.
        addv
            The insertion mode.

        See Also
        --------
        globalToNatural, petsc.DMDANaturalToGlobalBegin
        petsc.DMDANaturalToGlobalEnd

        """
        cdef PetscInsertMode im = insertmode(addv)
        CHKERR(DMDANaturalToGlobalBegin(self.dm, vn.vec, im, vg.vec))
        CHKERR(DMDANaturalToGlobalEnd(self.dm, vn.vec, im, vg.vec))

    #

    def getAO(self) -> AO:
        """Return the application ordering context for a distributed array.

        Collective.

        The returned `AO` maps to the natural grid ordering that would be
        used for the `DMDA` if only 1 processor were employed (ordering most
        rapidly in the x-dimension, then y, then z). Multiple degrees of
        freedom are numbered for each node (rather than 1 component for the
        whole grid, then the next component, etc.).

        See Also
        --------
        petsc.DMDAGetAO

        """
        cdef AO ao = AO()
        CHKERR(DMDAGetAO(self.dm, &ao.ao))
        CHKERR(PetscINCREF(ao.obj))
        return ao

    def getScatter(self) -> tuple[Scatter, Scatter]:
        """Return the global-to-local, and local-to-local scatter contexts.

        Collective.

        See Also
        --------
        petsc.DMDAGetScatter

        """
        cdef Scatter l2g = Scatter()
        cdef Scatter g2l = Scatter()
        CHKERR(DMDAGetScatter(self.dm, &l2g.sct, &g2l.sct))
        CHKERR(PetscINCREF(l2g.obj))
        CHKERR(PetscINCREF(g2l.obj))
        return (l2g, g2l)

    #

    def setRefinementFactor(
        self,
        refine_x: int = 2,
        refine_y: int = 2,
        refine_z: int = 2) -> None:
        """Set the ratios for the DMDA grid refinement.

        Logically collective.

        Parameters
        ----------
        refine_x
            Ratio of fine grid to coarse in x dimension.
        refine_y
            Ratio of fine grid to coarse in y dimension.
        refine_z
            Ratio of fine grid to coarse in z dimension.

        See Also
        --------
        getRefinementFactor, petsc.DMDASetRefinementFactor

        """
        cdef PetscInt refine[3]
        refine[0] = asInt(refine_x)
        refine[1] = asInt(refine_y)
        refine[2] = asInt(refine_z)
        CHKERR(DMDASetRefinementFactor(self.dm,
                                       refine[0],
                                       refine[1],
                                       refine[2]))

    def getRefinementFactor(self) -> tuple[int, ...]:
        """Return the ratios that the DMDA grid is refined in each dimension.

        Not collective.

        See Also
        --------
        setRefinementFactor, petsc.DMDAGetRefinementFactor

        """
        cdef PetscInt dim = 0, refine[3]
        CHKERR(DMDAGetDim(self.dm, &dim))
        CHKERR(DMDAGetRefinementFactor(self.dm,
                                       &refine[0],
                                       &refine[1],
                                       &refine[2]))
        return tuple([toInt(refine[i]) for 0 <= i < dim])

    def setInterpolationType(self, interp_type: InterpolationType) -> None:
        """Set the type of interpolation.

        Logically collective.

        You should call this on the coarser of the two DMDAs you pass to
        `DM.createInterpolation`.

        Parameters
        ----------
        interp_type
            The interpolation type.

        See Also
        --------
        getInterpolationType, petsc.DMDASetInterpolationType

        """
        cdef PetscDMDAInterpolationType ival = dainterpolationtype(interp_type)
        CHKERR(DMDASetInterpolationType(self.dm, ival))

    def getInterpolationType(self) -> InterpolationType:
        """Return the type of interpolation.

        Not collective.

        See Also
        --------
        setInterpolationType, petsc.DMDAGetInterpolationType

        """
        cdef PetscDMDAInterpolationType ival = DMDA_INTERPOLATION_Q0
        CHKERR(DMDAGetInterpolationType(self.dm, &ival))
        return <long>ival

    #

    def setElementType(self, elem_type: ElementType | str) -> None:
        """Set the element type to be returned by `getElements`.

        Not collective.

        See Also
        --------
        getElementType, petsc.DMDASetElementType

        """
        cdef PetscDMDAElementType ival = daelementtype(elem_type)
        CHKERR(DMDASetElementType(self.dm, ival))

    # FIXME: Return type
    def getElementType(self) -> ElementType:
        """Return the element type to be returned by `getElements`.

        Not collective.

        See Also
        --------
        setElementType, petsc.DMDAGetElementType

        """
        cdef PetscDMDAElementType ival = DMDA_ELEMENT_Q1
        CHKERR(DMDAGetElementType(self.dm, &ival))
        return <long>ival

    def getElements(self, elem_type: ElementType | None = None) -> ArrayInt:
        """Return an array containing the indices of all the local elements.

        Not collective.

        The elements are in local coordinates.

        Each process uniquely owns a subset of the elements. That is, no
        element is owned by two or more processes.

        Parameters
        ----------
        elem_type
            The element type.

        See Also
        --------
        petsc.DMDAGetElements

        """
        cdef PetscInt dim=0
        cdef PetscDMDAElementType etype
        cdef PetscInt nel=0, nen=0
        cdef const PetscInt *elems=NULL
        cdef object elements
        CHKERR(DMDAGetDim(self.dm, &dim))
        if elem_type is not None:
            etype = daelementtype(elem_type)
            CHKERR(DMDASetElementType(self.dm, etype))
        try:
            CHKERR(DMDAGetElements(self.dm, &nel, &nen, &elems))
            elements = array_i(nel*nen, elems)
            elements.shape = (toInt(nel), toInt(nen))
        finally:
            CHKERR(DMDARestoreElements(self.dm, &nel, &nen, &elems))
        return elements

    #

    property dim:
        """The grid dimension."""
        def __get__(self) -> int:
            return self.getDim()

    property dof:
        """The number of DOFs associated with each stratum of the grid."""
        def __get__(self) -> int:
            return self.getDof()

    property sizes:
        """The global dimension."""
        def __get__(self) -> tuple[int, ...]:
            return self.getSizes()

    property proc_sizes:
        """The number of processes in each dimension in the global decomposition."""
        def __get__(self) -> tuple[int, ...]:
            return self.getProcSizes()

    property boundary_type:
        """Boundary types in each dimension."""
        def __get__(self) -> tuple[DM.BoundaryType, ...]:
            return self.getBoundaryType()

    property stencil:
        """Stencil type and width."""
        def __get__(self) -> tuple[StencilType, int]:
            return self.getStencil()

    property stencil_type:
        """Stencil type."""
        def __get__(self) -> str:
            return self.getStencilType()

    property stencil_width:
        """Elementwise stencil width."""
        def __get__(self) -> int:
            return self.getStencilWidth()

    property ranges:
        """Ranges of the local region in each dimension."""
        def __get__(self) -> tuple[tuple[int, int], ...]:
            return self.getRanges()

    property ghost_ranges:
        """Ranges of local region, including ghost nodes."""
        def __get__(self) -> tuple[tuple[int, int], ...]:
            return self.getGhostRanges()

    property corners:
        """The lower left corner and size of local region in each dimension."""
        def __get__(self) -> tuple[tuple[int, ...], tuple[int, ...]]:
            return self.getCorners()

    property ghost_corners:
        """The lower left corner and size of local region in each dimension."""
        def __get__(self) -> tuple[tuple[int, ...], tuple[int, ...]]:
            return self.getGhostCorners()

    # backward compatibility
    createNaturalVector = createNaturalVec


# backward compatibility alias
DA = DMDA

# --------------------------------------------------------------------

del DMDAStencilType
del DMDAInterpolationType
del DMDAElementType

# --------------------------------------------------------------------



#!/usr/bin/env python3
"""Run Cython with custom options."""

import os
import sys

appdir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, appdir)

import cyautodoc  # noqa: F401,E402
from Cython.Compiler.Main import main as cython_main  # noqa: E402


def cythonize(args=None):
    """Run `cython --3str --cleanup 3 <args>...`."""
    if args is None:
        argv = sys.argv[:]
    else:
        argv = [os.path.abspath(__file__)] + list(args)

    if '--cleanup' not in argv:
        argv[1:1] = ['--cleanup', '3']
    if '--3str' not in argv:
        argv[1:1] = ['--3str']

    cwd = os.getcwd()
    sys_argv = sys.argv[:]
    try:
        sys.argv[:] = argv
        cython_main(command_line=1)
        return 0
    except SystemExit as exc:
        return exc.code
    finally:
        os.chdir(cwd)
        sys.argv[:] = sys_argv


def main():
    """Entry-point to run Cython with custom options."""
    args = sys.argv[1:]
    if not args:
        topdir = os.path.dirname(appdir)
        srcdir = os.path.join(topdir, 'src')
        source = os.path.join('petsc4py', 'PETSc.pyx')
        target = os.path.join('petsc4py', 'PETSc.c')
        args += ['--working', srcdir]
        args += [source, '--output-file', target]
    sys.exit(cythonize(args))


if __name__ == '__main__':
    main()



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com

# --------------------------------------------------------------------

"""
Extension modules for different PETSc configurations.

PETSc can be configured with different options (eg. debug/optimized,
single/double precisionm, C/C++ compilers, external packages). Each
configuration variant is associated to a name, frequently available as
an environmental variable named ``PETSC_ARCH``.

This package holds all the available variants of the PETSc
extension module built against specific PETSc configurations. It also
provides a convenience function using of the ``importlib`` module
for easily importing any of the available extension modules depending
on the value of a user-provided configuration name, the ``PETSC_ARCH``
environmental variable, or a configuration file.

Python implementations of the Python aware PETSc types are also
exposed in submodules here.
"""

# --------------------------------------------------------------------


def ImportPETSc(arch=None):
    """
    Import the PETSc extension module for a given configuration name.
    """
    path, arch = getPathArchPETSc(arch)
    return Import('petsc4py', 'PETSc', path, arch)


def getPathArchPETSc(arch=None):
    """
    Undocumented.
    """
    import os

    path = os.path.abspath(os.path.dirname(__file__))
    rcvar, rcfile = 'PETSC_ARCH', 'petsc.cfg'
    path, arch = getPathArch(path, arch, rcvar, rcfile)
    return (path, arch)


# --------------------------------------------------------------------


def Import(pkg, name, path, arch):
    """
    Import helper for PETSc-based extension modules.
    """
    import os
    import sys
    import warnings

    try:
        import importlib.machinery
        import importlib.util
    except ImportError:
        importlib = None
        import imp

    def get_ext_suffix():
        if importlib:
            return importlib.machinery.EXTENSION_SUFFIXES[0]
        return imp.get_suffixes()[0][0]

    def import_module(pkg, name, path, arch):
        fullname = f'{pkg}.{name}'
        pathlist = [os.path.join(path, arch)]
        if importlib:
            finder = importlib.machinery.PathFinder()
            spec = finder.find_spec(fullname, pathlist)
            module = importlib.util.module_from_spec(spec)
            sys.modules[fullname] = module
            spec.loader.exec_module(module)
            return module
        f, fn, info = imp.find_module(name, pathlist)
        with f:
            return imp.load_module(fullname, f, fn, info)

    # test if extension module was already imported
    module = sys.modules.get(f'{pkg}.{name}')
    filename = getattr(module, '__file__', '')
    if filename.endswith(get_ext_suffix()):
        # if 'arch' is None, do nothing; otherwise this
        # call may be invalid if extension module for
        # other 'arch' has been already imported.
        if arch is not None and arch != module.__arch__:
            raise ImportError('%s already imported' % module)
        return module

    # silence annoying Cython warning
    warnings.filterwarnings('ignore', message='numpy.dtype size changed')
    warnings.filterwarnings('ignore', message='numpy.ndarray size changed')
    # import extension module from 'path/arch' directory
    module = import_module(pkg, name, path, arch)
    module.__arch__ = arch  # save arch value
    setattr(sys.modules[pkg], name, module)
    return module


def getPathArch(path, arch, rcvar='PETSC_ARCH', rcfile='petsc.cfg'):
    """
    Undocumented.
    """
    import os
    import warnings

    # path
    if not path:
        path = '.'
    elif os.path.isfile(path):
        path = os.path.dirname(path)
    elif not os.path.isdir(path):
        raise ValueError("invalid path: '%s'" % path)
    # arch
    if arch is not None:
        if not isinstance(arch, str):
            raise TypeError('arch argument must be string')
        if not os.path.isdir(os.path.join(path, arch)):
            raise TypeError("invalid arch value: '%s'" % arch)
        return (path, arch)

    # helper function
    def arch_list(arch):
        arch = arch.strip().split(os.path.pathsep)
        arch = [a.strip() for a in arch if a]
        return [a for a in arch if a]

    # try to get arch from the environment
    arch_env = arch_list(os.environ.get(rcvar, ''))
    for arch in arch_env:
        if os.path.isdir(os.path.join(path, arch)):
            return (path, arch)
    # configuration file
    if not os.path.isfile(rcfile):
        rcfile = os.path.join(path, rcfile)
        if not os.path.isfile(rcfile):
            # now point to continue
            return (path, '')

    # helper function
    def parse_rc(rcfile):
        with open(rcfile) as f:
            rcdata = f.read()
        lines = [ln.strip() for ln in rcdata.splitlines()]
        lines = [ln for ln in lines if not ln.startswith('#')]
        entries = [ln.split('=') for ln in lines if ln]
        entries = [(k.strip(), v.strip()) for k, v in entries]
        return dict(entries)

    # try to get arch from data in config file
    configrc = parse_rc(rcfile)
    arch_cfg = arch_list(configrc.get(rcvar, ''))
    for arch in arch_cfg:
        if arch.startswith('%(') and arch.endswith(')s'):
            arch = arch % os.environ
        if os.path.isdir(os.path.join(path, arch)):
            if arch_env:
                warnings.warn(
                    f"ignored arch: '{os.path.pathsep.join(arch_env)}', using: '{arch}'",
                    stacklevel=2,
                )
            return (path, arch)
    # nothing good found
    return (path, '')


def getInitArgs(args):
    """
    Undocumented.
    """
    import shlex
    import sys

    if args is None:
        args = []
    elif isinstance(args, str):
        args = shlex.split(args)
    else:
        args = [str(a) for a in args]
        args = [a for a in args if a]
    if args and args[0].startswith('-'):
        sys_argv = getattr(sys, 'argv', None)
        sys_exec = getattr(sys, 'executable', 'python')
        if sys_argv and sys_argv[0] and sys_argv[0] != '-c':
            prog_name = sys_argv[0]
        else:
            prog_name = sys_exec
        args.insert(0, prog_name)
    return args


# --------------------------------------------------------------------
